from mflux.models.common.lora.mapping.lora_mapping import LoRAMapping, LoRATarget
from mflux.models.common.lora.mapping.lora_transforms import LoraTransforms


class Flux2LoRAMapping(LoRAMapping):
    @staticmethod
    def get_mapping() -> list[LoRATarget]:
        targets: list[LoRATarget] = []

        targets.extend(Flux2LoRAMapping._get_global_targets())

        # Standard (explicit) naming conventions
        targets.extend(Flux2LoRAMapping._get_double_stream_block_targets())
        targets.extend(Flux2LoRAMapping._get_single_stream_block_targets())

        # BFL / ComfyUI-style naming conventions (qkv split, etc.)
        targets.extend(Flux2LoRAMapping._get_bfl_double_stream_block_targets())
        targets.extend(Flux2LoRAMapping._get_bfl_single_stream_block_targets())

        return targets

    @staticmethod
    def _get_global_targets() -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path="x_embedder",
                possible_up_patterns=[
                    "x_embedder.lora_B.weight",
                    "x_embedder.lora_up.weight",
                    "transformer.x_embedder.lora_B.weight",
                    "transformer.x_embedder.lora_up.weight",
                    "diffusion_model.x_embedder.lora_B.weight",
                    "diffusion_model.x_embedder.lora_up.weight",
                    "base_model.model.img_in.lora_B.weight",
                    "base_model.model.img_in.lora_up.weight",
                ],
                possible_down_patterns=[
                    "x_embedder.lora_A.weight",
                    "x_embedder.lora_down.weight",
                    "transformer.x_embedder.lora_A.weight",
                    "transformer.x_embedder.lora_down.weight",
                    "diffusion_model.x_embedder.lora_A.weight",
                    "diffusion_model.x_embedder.lora_down.weight",
                    "base_model.model.img_in.lora_A.weight",
                    "base_model.model.img_in.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "x_embedder.alpha",
                    "transformer.x_embedder.alpha",
                    "diffusion_model.x_embedder.alpha",
                    "base_model.model.img_in.alpha",
                ],
            ),
            LoRATarget(
                model_path="context_embedder",
                possible_up_patterns=[
                    "context_embedder.lora_B.weight",
                    "context_embedder.lora_up.weight",
                    "transformer.context_embedder.lora_B.weight",
                    "transformer.context_embedder.lora_up.weight",
                    "diffusion_model.context_embedder.lora_B.weight",
                    "diffusion_model.context_embedder.lora_up.weight",
                    "base_model.model.txt_in.lora_B.weight",
                    "base_model.model.txt_in.lora_up.weight",
                ],
                possible_down_patterns=[
                    "context_embedder.lora_A.weight",
                    "context_embedder.lora_down.weight",
                    "transformer.context_embedder.lora_A.weight",
                    "transformer.context_embedder.lora_down.weight",
                    "diffusion_model.context_embedder.lora_A.weight",
                    "diffusion_model.context_embedder.lora_down.weight",
                    "base_model.model.txt_in.lora_A.weight",
                    "base_model.model.txt_in.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "context_embedder.alpha",
                    "transformer.context_embedder.alpha",
                    "diffusion_model.context_embedder.alpha",
                    "base_model.model.txt_in.alpha",
                ],
            ),
            LoRATarget(
                model_path="time_guidance_embed.linear_1",
                possible_up_patterns=[
                    "time_guidance_embed.linear_1.lora_B.weight",
                    "time_guidance_embed.linear_1.lora_up.weight",
                    "transformer.time_guidance_embed.linear_1.lora_B.weight",
                    "transformer.time_guidance_embed.linear_1.lora_up.weight",
                    "diffusion_model.time_guidance_embed.linear_1.lora_B.weight",
                    "diffusion_model.time_guidance_embed.linear_1.lora_up.weight",
                    "base_model.model.time_in.in_layer.lora_B.weight",
                    "base_model.model.time_in.in_layer.lora_up.weight",
                ],
                possible_down_patterns=[
                    "time_guidance_embed.linear_1.lora_A.weight",
                    "time_guidance_embed.linear_1.lora_down.weight",
                    "transformer.time_guidance_embed.linear_1.lora_A.weight",
                    "transformer.time_guidance_embed.linear_1.lora_down.weight",
                    "diffusion_model.time_guidance_embed.linear_1.lora_A.weight",
                    "diffusion_model.time_guidance_embed.linear_1.lora_down.weight",
                    "base_model.model.time_in.in_layer.lora_A.weight",
                    "base_model.model.time_in.in_layer.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "time_guidance_embed.linear_1.alpha",
                    "transformer.time_guidance_embed.linear_1.alpha",
                    "diffusion_model.time_guidance_embed.linear_1.alpha",
                    "base_model.model.time_in.in_layer.alpha",
                ],
            ),
            LoRATarget(
                model_path="time_guidance_embed.linear_2",
                possible_up_patterns=[
                    "time_guidance_embed.linear_2.lora_B.weight",
                    "time_guidance_embed.linear_2.lora_up.weight",
                    "transformer.time_guidance_embed.linear_2.lora_B.weight",
                    "transformer.time_guidance_embed.linear_2.lora_up.weight",
                    "diffusion_model.time_guidance_embed.linear_2.lora_B.weight",
                    "diffusion_model.time_guidance_embed.linear_2.lora_up.weight",
                    "base_model.model.time_in.out_layer.lora_B.weight",
                    "base_model.model.time_in.out_layer.lora_up.weight",
                ],
                possible_down_patterns=[
                    "time_guidance_embed.linear_2.lora_A.weight",
                    "time_guidance_embed.linear_2.lora_down.weight",
                    "transformer.time_guidance_embed.linear_2.lora_A.weight",
                    "transformer.time_guidance_embed.linear_2.lora_down.weight",
                    "diffusion_model.time_guidance_embed.linear_2.lora_A.weight",
                    "diffusion_model.time_guidance_embed.linear_2.lora_down.weight",
                    "base_model.model.time_in.out_layer.lora_A.weight",
                    "base_model.model.time_in.out_layer.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "time_guidance_embed.linear_2.alpha",
                    "transformer.time_guidance_embed.linear_2.alpha",
                    "diffusion_model.time_guidance_embed.linear_2.alpha",
                    "base_model.model.time_in.out_layer.alpha",
                ],
            ),
            # Optional guidance embeddings (only present if guidance_embeds=True)
            LoRATarget(
                model_path="time_guidance_embed.guidance_linear_1",
                possible_up_patterns=[
                    "time_guidance_embed.guidance_linear_1.lora_B.weight",
                    "time_guidance_embed.guidance_linear_1.lora_up.weight",
                    "transformer.time_guidance_embed.guidance_linear_1.lora_B.weight",
                    "transformer.time_guidance_embed.guidance_linear_1.lora_up.weight",
                    "diffusion_model.time_guidance_embed.guidance_linear_1.lora_B.weight",
                    "diffusion_model.time_guidance_embed.guidance_linear_1.lora_up.weight",
                ],
                possible_down_patterns=[
                    "time_guidance_embed.guidance_linear_1.lora_A.weight",
                    "time_guidance_embed.guidance_linear_1.lora_down.weight",
                    "transformer.time_guidance_embed.guidance_linear_1.lora_A.weight",
                    "transformer.time_guidance_embed.guidance_linear_1.lora_down.weight",
                    "diffusion_model.time_guidance_embed.guidance_linear_1.lora_A.weight",
                    "diffusion_model.time_guidance_embed.guidance_linear_1.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "time_guidance_embed.guidance_linear_1.alpha",
                    "transformer.time_guidance_embed.guidance_linear_1.alpha",
                    "diffusion_model.time_guidance_embed.guidance_linear_1.alpha",
                ],
            ),
            LoRATarget(
                model_path="time_guidance_embed.guidance_linear_2",
                possible_up_patterns=[
                    "time_guidance_embed.guidance_linear_2.lora_B.weight",
                    "time_guidance_embed.guidance_linear_2.lora_up.weight",
                    "transformer.time_guidance_embed.guidance_linear_2.lora_B.weight",
                    "transformer.time_guidance_embed.guidance_linear_2.lora_up.weight",
                    "diffusion_model.time_guidance_embed.guidance_linear_2.lora_B.weight",
                    "diffusion_model.time_guidance_embed.guidance_linear_2.lora_up.weight",
                ],
                possible_down_patterns=[
                    "time_guidance_embed.guidance_linear_2.lora_A.weight",
                    "time_guidance_embed.guidance_linear_2.lora_down.weight",
                    "transformer.time_guidance_embed.guidance_linear_2.lora_A.weight",
                    "transformer.time_guidance_embed.guidance_linear_2.lora_down.weight",
                    "diffusion_model.time_guidance_embed.guidance_linear_2.lora_A.weight",
                    "diffusion_model.time_guidance_embed.guidance_linear_2.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "time_guidance_embed.guidance_linear_2.alpha",
                    "transformer.time_guidance_embed.guidance_linear_2.alpha",
                    "diffusion_model.time_guidance_embed.guidance_linear_2.alpha",
                ],
            ),
            LoRATarget(
                model_path="double_stream_modulation_img.linear",
                possible_up_patterns=[
                    "double_stream_modulation_img.linear.lora_B.weight",
                    "double_stream_modulation_img.linear.lora_up.weight",
                    "transformer.double_stream_modulation_img.linear.lora_B.weight",
                    "transformer.double_stream_modulation_img.linear.lora_up.weight",
                    "diffusion_model.double_stream_modulation_img.linear.lora_B.weight",
                    "diffusion_model.double_stream_modulation_img.linear.lora_up.weight",
                    "base_model.model.double_stream_modulation_img.lin.lora_B.weight",
                    "base_model.model.double_stream_modulation_img.lin.lora_up.weight",
                ],
                possible_down_patterns=[
                    "double_stream_modulation_img.linear.lora_A.weight",
                    "double_stream_modulation_img.linear.lora_down.weight",
                    "transformer.double_stream_modulation_img.linear.lora_A.weight",
                    "transformer.double_stream_modulation_img.linear.lora_down.weight",
                    "diffusion_model.double_stream_modulation_img.linear.lora_A.weight",
                    "diffusion_model.double_stream_modulation_img.linear.lora_down.weight",
                    "base_model.model.double_stream_modulation_img.lin.lora_A.weight",
                    "base_model.model.double_stream_modulation_img.lin.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "double_stream_modulation_img.linear.alpha",
                    "transformer.double_stream_modulation_img.linear.alpha",
                    "diffusion_model.double_stream_modulation_img.linear.alpha",
                    "base_model.model.double_stream_modulation_img.lin.alpha",
                ],
            ),
            LoRATarget(
                model_path="double_stream_modulation_txt.linear",
                possible_up_patterns=[
                    "double_stream_modulation_txt.linear.lora_B.weight",
                    "double_stream_modulation_txt.linear.lora_up.weight",
                    "transformer.double_stream_modulation_txt.linear.lora_B.weight",
                    "transformer.double_stream_modulation_txt.linear.lora_up.weight",
                    "diffusion_model.double_stream_modulation_txt.linear.lora_B.weight",
                    "diffusion_model.double_stream_modulation_txt.linear.lora_up.weight",
                    "base_model.model.double_stream_modulation_txt.lin.lora_B.weight",
                    "base_model.model.double_stream_modulation_txt.lin.lora_up.weight",
                ],
                possible_down_patterns=[
                    "double_stream_modulation_txt.linear.lora_A.weight",
                    "double_stream_modulation_txt.linear.lora_down.weight",
                    "transformer.double_stream_modulation_txt.linear.lora_A.weight",
                    "transformer.double_stream_modulation_txt.linear.lora_down.weight",
                    "diffusion_model.double_stream_modulation_txt.linear.lora_A.weight",
                    "diffusion_model.double_stream_modulation_txt.linear.lora_down.weight",
                    "base_model.model.double_stream_modulation_txt.lin.lora_A.weight",
                    "base_model.model.double_stream_modulation_txt.lin.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "double_stream_modulation_txt.linear.alpha",
                    "transformer.double_stream_modulation_txt.linear.alpha",
                    "diffusion_model.double_stream_modulation_txt.linear.alpha",
                    "base_model.model.double_stream_modulation_txt.lin.alpha",
                ],
            ),
            LoRATarget(
                model_path="single_stream_modulation.linear",
                possible_up_patterns=[
                    "single_stream_modulation.linear.lora_B.weight",
                    "single_stream_modulation.linear.lora_up.weight",
                    "transformer.single_stream_modulation.linear.lora_B.weight",
                    "transformer.single_stream_modulation.linear.lora_up.weight",
                    "diffusion_model.single_stream_modulation.linear.lora_B.weight",
                    "diffusion_model.single_stream_modulation.linear.lora_up.weight",
                    "base_model.model.single_stream_modulation.lin.lora_B.weight",
                    "base_model.model.single_stream_modulation.lin.lora_up.weight",
                ],
                possible_down_patterns=[
                    "single_stream_modulation.linear.lora_A.weight",
                    "single_stream_modulation.linear.lora_down.weight",
                    "transformer.single_stream_modulation.linear.lora_A.weight",
                    "transformer.single_stream_modulation.linear.lora_down.weight",
                    "diffusion_model.single_stream_modulation.linear.lora_A.weight",
                    "diffusion_model.single_stream_modulation.linear.lora_down.weight",
                    "base_model.model.single_stream_modulation.lin.lora_A.weight",
                    "base_model.model.single_stream_modulation.lin.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "single_stream_modulation.linear.alpha",
                    "transformer.single_stream_modulation.linear.alpha",
                    "diffusion_model.single_stream_modulation.linear.alpha",
                    "base_model.model.single_stream_modulation.lin.alpha",
                ],
            ),
            LoRATarget(
                model_path="norm_out.linear",
                possible_up_patterns=[
                    "norm_out.linear.lora_B.weight",
                    "norm_out.linear.lora_up.weight",
                    "transformer.norm_out.linear.lora_B.weight",
                    "transformer.norm_out.linear.lora_up.weight",
                    "diffusion_model.norm_out.linear.lora_B.weight",
                    "diffusion_model.norm_out.linear.lora_up.weight",
                ],
                possible_down_patterns=[
                    "norm_out.linear.lora_A.weight",
                    "norm_out.linear.lora_down.weight",
                    "transformer.norm_out.linear.lora_A.weight",
                    "transformer.norm_out.linear.lora_down.weight",
                    "diffusion_model.norm_out.linear.lora_A.weight",
                    "diffusion_model.norm_out.linear.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "norm_out.linear.alpha",
                    "transformer.norm_out.linear.alpha",
                    "diffusion_model.norm_out.linear.alpha",
                ],
            ),
            LoRATarget(
                model_path="proj_out",
                possible_up_patterns=[
                    "proj_out.lora_B.weight",
                    "proj_out.lora_up.weight",
                    "transformer.proj_out.lora_B.weight",
                    "transformer.proj_out.lora_up.weight",
                    "diffusion_model.proj_out.lora_B.weight",
                    "diffusion_model.proj_out.lora_up.weight",
                    "base_model.model.final_layer.linear.lora_B.weight",
                    "base_model.model.final_layer.linear.lora_up.weight",
                ],
                possible_down_patterns=[
                    "proj_out.lora_A.weight",
                    "proj_out.lora_down.weight",
                    "transformer.proj_out.lora_A.weight",
                    "transformer.proj_out.lora_down.weight",
                    "diffusion_model.proj_out.lora_A.weight",
                    "diffusion_model.proj_out.lora_down.weight",
                    "base_model.model.final_layer.linear.lora_A.weight",
                    "base_model.model.final_layer.linear.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "proj_out.alpha",
                    "transformer.proj_out.alpha",
                    "diffusion_model.proj_out.alpha",
                    "base_model.model.final_layer.linear.alpha",
                ],
            ),
        ]

    @staticmethod
    def _get_double_stream_block_targets() -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_q",
                possible_up_patterns=[
                    "transformer_blocks.{block}.attn.to_q.lora_B.weight",
                    "transformer_blocks.{block}.attn.to_q.lora_B.default.weight",
                    "transformer_blocks.{block}.attn.to_q.lora_up.weight",
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_q.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_q.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_q.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_q.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.attn.to_q.lora_A.weight",
                    "transformer_blocks.{block}.attn.to_q.lora_A.default.weight",
                    "transformer_blocks.{block}.attn.to_q.lora_down.weight",
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_q.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_q.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_q.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_q.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_q.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.attn.to_q.alpha",
                    "transformer.transformer_blocks.{block}.attn.to_q.alpha",
                    "diffusion_model.transformer_blocks.{block}.attn.to_q.alpha",
                    "lora_unet_transformer_blocks_{block}_attn_to_q.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_k",
                possible_up_patterns=[
                    "transformer_blocks.{block}.attn.to_k.lora_B.weight",
                    "transformer_blocks.{block}.attn.to_k.lora_B.default.weight",
                    "transformer_blocks.{block}.attn.to_k.lora_up.weight",
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_k.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_k.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_k.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_k.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.attn.to_k.lora_A.weight",
                    "transformer_blocks.{block}.attn.to_k.lora_A.default.weight",
                    "transformer_blocks.{block}.attn.to_k.lora_down.weight",
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_k.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_k.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_k.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_k.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_k.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.attn.to_k.alpha",
                    "transformer.transformer_blocks.{block}.attn.to_k.alpha",
                    "diffusion_model.transformer_blocks.{block}.attn.to_k.alpha",
                    "lora_unet_transformer_blocks_{block}_attn_to_k.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_v",
                possible_up_patterns=[
                    "transformer_blocks.{block}.attn.to_v.lora_B.weight",
                    "transformer_blocks.{block}.attn.to_v.lora_B.default.weight",
                    "transformer_blocks.{block}.attn.to_v.lora_up.weight",
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_v.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_v.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_v.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_v.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.attn.to_v.lora_A.weight",
                    "transformer_blocks.{block}.attn.to_v.lora_A.default.weight",
                    "transformer_blocks.{block}.attn.to_v.lora_down.weight",
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_v.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_v.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_v.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_v.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_v.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.attn.to_v.alpha",
                    "transformer.transformer_blocks.{block}.attn.to_v.alpha",
                    "diffusion_model.transformer_blocks.{block}.attn.to_v.alpha",
                    "lora_unet_transformer_blocks_{block}_attn_to_v.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_out",
                possible_up_patterns=[
                    "transformer_blocks.{block}.attn.to_out.lora_B.weight",
                    "transformer_blocks.{block}.attn.to_out.lora_B.default.weight",
                    "transformer_blocks.{block}.attn.to_out.lora_up.weight",
                    "transformer_blocks.{block}.attn.to_out.0.lora_B.weight",  # HF-style
                    "transformer_blocks.{block}.attn.to_out.0.lora_B.default.weight",  # HF-style
                    "transformer_blocks.{block}.attn.to_out.0.lora_up.weight",  # HF-style
                    "transformer.transformer_blocks.{block}.attn.to_out.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_out.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_out.lora_up.weight",
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_B.weight",  # HF-style
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_B.default.weight",  # HF-style
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_up.weight",  # HF-style
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.0.lora_B.weight",  # HF-style
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.0.lora_B.default.weight",  # HF-style
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.0.lora_up.weight",  # HF-style
                    "lora_unet_transformer_blocks_{block}_attn_to_out.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_out_0.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.attn.to_out.lora_A.weight",
                    "transformer_blocks.{block}.attn.to_out.lora_A.default.weight",
                    "transformer_blocks.{block}.attn.to_out.lora_down.weight",
                    "transformer_blocks.{block}.attn.to_out.0.lora_A.weight",  # HF-style
                    "transformer_blocks.{block}.attn.to_out.0.lora_A.default.weight",  # HF-style
                    "transformer_blocks.{block}.attn.to_out.0.lora_down.weight",  # HF-style
                    "transformer.transformer_blocks.{block}.attn.to_out.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_out.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_out.lora_down.weight",
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_A.weight",  # HF-style
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_A.default.weight",  # HF-style
                    "transformer.transformer_blocks.{block}.attn.to_out.0.lora_down.weight",  # HF-style
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.0.lora_A.weight",  # HF-style
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.0.lora_A.default.weight",  # HF-style
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.0.lora_down.weight",  # HF-style
                    "lora_unet_transformer_blocks_{block}_attn_to_out.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_out_0.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.attn.to_out.alpha",
                    "transformer_blocks.{block}.attn.to_out.0.alpha",
                    "transformer.transformer_blocks.{block}.attn.to_out.alpha",
                    "transformer.transformer_blocks.{block}.attn.to_out.0.alpha",
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.alpha",
                    "diffusion_model.transformer_blocks.{block}.attn.to_out.0.alpha",
                    "lora_unet_transformer_blocks_{block}_attn_to_out.alpha",
                    "lora_unet_transformer_blocks_{block}_attn_to_out_0.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_q_proj",
                possible_up_patterns=[
                    "transformer_blocks.{block}.attn.add_q_proj.lora_B.weight",
                    "transformer_blocks.{block}.attn.add_q_proj.lora_B.default.weight",
                    "transformer_blocks.{block}.attn.add_q_proj.lora_up.weight",
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_q_proj.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_q_proj.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_q_proj.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_attn_add_q_proj.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.attn.add_q_proj.lora_A.weight",
                    "transformer_blocks.{block}.attn.add_q_proj.lora_A.default.weight",
                    "transformer_blocks.{block}.attn.add_q_proj.lora_down.weight",
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_q_proj.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_q_proj.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_q_proj.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_attn_add_q_proj.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.attn.add_q_proj.alpha",
                    "transformer.transformer_blocks.{block}.attn.add_q_proj.alpha",
                    "diffusion_model.transformer_blocks.{block}.attn.add_q_proj.alpha",
                    "lora_unet_transformer_blocks_{block}_attn_add_q_proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_k_proj",
                possible_up_patterns=[
                    "transformer_blocks.{block}.attn.add_k_proj.lora_B.weight",
                    "transformer_blocks.{block}.attn.add_k_proj.lora_B.default.weight",
                    "transformer_blocks.{block}.attn.add_k_proj.lora_up.weight",
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_k_proj.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_k_proj.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_k_proj.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_attn_add_k_proj.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.attn.add_k_proj.lora_A.weight",
                    "transformer_blocks.{block}.attn.add_k_proj.lora_A.default.weight",
                    "transformer_blocks.{block}.attn.add_k_proj.lora_down.weight",
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_k_proj.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_k_proj.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_k_proj.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_attn_add_k_proj.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.attn.add_k_proj.alpha",
                    "transformer.transformer_blocks.{block}.attn.add_k_proj.alpha",
                    "diffusion_model.transformer_blocks.{block}.attn.add_k_proj.alpha",
                    "lora_unet_transformer_blocks_{block}_attn_add_k_proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_v_proj",
                possible_up_patterns=[
                    "transformer_blocks.{block}.attn.add_v_proj.lora_B.weight",
                    "transformer_blocks.{block}.attn.add_v_proj.lora_B.default.weight",
                    "transformer_blocks.{block}.attn.add_v_proj.lora_up.weight",
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_v_proj.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_v_proj.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_v_proj.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_attn_add_v_proj.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.attn.add_v_proj.lora_A.weight",
                    "transformer_blocks.{block}.attn.add_v_proj.lora_A.default.weight",
                    "transformer_blocks.{block}.attn.add_v_proj.lora_down.weight",
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_v_proj.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_v_proj.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.add_v_proj.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_attn_add_v_proj.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.attn.add_v_proj.alpha",
                    "transformer.transformer_blocks.{block}.attn.add_v_proj.alpha",
                    "diffusion_model.transformer_blocks.{block}.attn.add_v_proj.alpha",
                    "lora_unet_transformer_blocks_{block}_attn_add_v_proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_add_out",
                possible_up_patterns=[
                    "transformer_blocks.{block}.attn.to_add_out.lora_B.weight",
                    "transformer_blocks.{block}.attn.to_add_out.lora_B.default.weight",
                    "transformer_blocks.{block}.attn.to_add_out.lora_up.weight",
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_B.weight",
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_add_out.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_add_out.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_add_out.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_add_out.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.attn.to_add_out.lora_A.weight",
                    "transformer_blocks.{block}.attn.to_add_out.lora_A.default.weight",
                    "transformer_blocks.{block}.attn.to_add_out.lora_down.weight",
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_A.weight",
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.attn.to_add_out.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_add_out.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_add_out.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.attn.to_add_out.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_attn_to_add_out.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.attn.to_add_out.alpha",
                    "transformer.transformer_blocks.{block}.attn.to_add_out.alpha",
                    "diffusion_model.transformer_blocks.{block}.attn.to_add_out.alpha",
                    "lora_unet_transformer_blocks_{block}_attn_to_add_out.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff.linear_in",
                possible_up_patterns=[
                    "transformer_blocks.{block}.ff.linear_in.lora_B.weight",
                    "transformer_blocks.{block}.ff.linear_in.lora_B.default.weight",
                    "transformer_blocks.{block}.ff.linear_in.lora_up.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_in.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_in.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_in.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_in.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_in.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_in.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_ff_linear_in.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.ff.linear_in.lora_A.weight",
                    "transformer_blocks.{block}.ff.linear_in.lora_A.default.weight",
                    "transformer_blocks.{block}.ff.linear_in.lora_down.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_in.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_in.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_in.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_in.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_in.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_in.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_ff_linear_in.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.ff.linear_in.alpha",
                    "transformer.transformer_blocks.{block}.ff.linear_in.alpha",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_in.alpha",
                    "lora_unet_transformer_blocks_{block}_ff_linear_in.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff.linear_out",
                possible_up_patterns=[
                    "transformer_blocks.{block}.ff.linear_out.lora_B.weight",
                    "transformer_blocks.{block}.ff.linear_out.lora_B.default.weight",
                    "transformer_blocks.{block}.ff.linear_out.lora_up.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_out.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_out.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_out.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_out.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_out.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_out.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_ff_linear_out.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.ff.linear_out.lora_A.weight",
                    "transformer_blocks.{block}.ff.linear_out.lora_A.default.weight",
                    "transformer_blocks.{block}.ff.linear_out.lora_down.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_out.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_out.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.ff.linear_out.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_out.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_out.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_out.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_ff_linear_out.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.ff.linear_out.alpha",
                    "transformer.transformer_blocks.{block}.ff.linear_out.alpha",
                    "diffusion_model.transformer_blocks.{block}.ff.linear_out.alpha",
                    "lora_unet_transformer_blocks_{block}_ff_linear_out.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff_context.linear_in",
                possible_up_patterns=[
                    "transformer_blocks.{block}.ff_context.linear_in.lora_B.weight",
                    "transformer_blocks.{block}.ff_context.linear_in.lora_B.default.weight",
                    "transformer_blocks.{block}.ff_context.linear_in.lora_up.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_in.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_in.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_in.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_in.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_in.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_in.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_ff_context_linear_in.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.ff_context.linear_in.lora_A.weight",
                    "transformer_blocks.{block}.ff_context.linear_in.lora_A.default.weight",
                    "transformer_blocks.{block}.ff_context.linear_in.lora_down.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_in.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_in.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_in.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_in.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_in.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_in.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_ff_context_linear_in.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.ff_context.linear_in.alpha",
                    "transformer.transformer_blocks.{block}.ff_context.linear_in.alpha",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_in.alpha",
                    "lora_unet_transformer_blocks_{block}_ff_context_linear_in.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff_context.linear_out",
                possible_up_patterns=[
                    "transformer_blocks.{block}.ff_context.linear_out.lora_B.weight",
                    "transformer_blocks.{block}.ff_context.linear_out.lora_B.default.weight",
                    "transformer_blocks.{block}.ff_context.linear_out.lora_up.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_out.lora_B.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_out.lora_B.default.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_out.lora_up.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_out.lora_B.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_out.lora_B.default.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_out.lora_up.weight",
                    "lora_unet_transformer_blocks_{block}_ff_context_linear_out.lora_up.weight",
                ],
                possible_down_patterns=[
                    "transformer_blocks.{block}.ff_context.linear_out.lora_A.weight",
                    "transformer_blocks.{block}.ff_context.linear_out.lora_A.default.weight",
                    "transformer_blocks.{block}.ff_context.linear_out.lora_down.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_out.lora_A.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_out.lora_A.default.weight",
                    "transformer.transformer_blocks.{block}.ff_context.linear_out.lora_down.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_out.lora_A.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_out.lora_A.default.weight",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_out.lora_down.weight",
                    "lora_unet_transformer_blocks_{block}_ff_context_linear_out.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "transformer_blocks.{block}.ff_context.linear_out.alpha",
                    "transformer.transformer_blocks.{block}.ff_context.linear_out.alpha",
                    "diffusion_model.transformer_blocks.{block}.ff_context.linear_out.alpha",
                    "lora_unet_transformer_blocks_{block}_ff_context_linear_out.alpha",
                ],
            ),
        ]

    @staticmethod
    def _get_single_stream_block_targets() -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_qkv_mlp_proj",
                possible_up_patterns=[
                    "single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_B.weight",
                    "single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_B.default.weight",
                    "single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_up.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_B.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_B.default.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_up.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_B.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_B.default.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_up.weight",
                    "lora_unet_single_transformer_blocks_{block}_attn_to_qkv_mlp_proj.lora_up.weight",
                ],
                possible_down_patterns=[
                    "single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_A.weight",
                    "single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_A.default.weight",
                    "single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_down.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_A.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_A.default.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_down.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_A.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_A.default.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.lora_down.weight",
                    "lora_unet_single_transformer_blocks_{block}_attn_to_qkv_mlp_proj.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.alpha",
                    "transformer.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.alpha",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_qkv_mlp_proj.alpha",
                    "lora_unet_single_transformer_blocks_{block}_attn_to_qkv_mlp_proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_out",
                possible_up_patterns=[
                    "single_transformer_blocks.{block}.attn.to_out.lora_B.weight",
                    "single_transformer_blocks.{block}.attn.to_out.lora_B.default.weight",
                    "single_transformer_blocks.{block}.attn.to_out.lora_up.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_out.lora_B.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_out.lora_B.default.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_out.lora_up.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_out.lora_B.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_out.lora_B.default.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_out.lora_up.weight",
                    "lora_unet_single_transformer_blocks_{block}_attn_to_out.lora_up.weight",
                ],
                possible_down_patterns=[
                    "single_transformer_blocks.{block}.attn.to_out.lora_A.weight",
                    "single_transformer_blocks.{block}.attn.to_out.lora_A.default.weight",
                    "single_transformer_blocks.{block}.attn.to_out.lora_down.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_out.lora_A.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_out.lora_A.default.weight",
                    "transformer.single_transformer_blocks.{block}.attn.to_out.lora_down.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_out.lora_A.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_out.lora_A.default.weight",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_out.lora_down.weight",
                    "lora_unet_single_transformer_blocks_{block}_attn_to_out.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "single_transformer_blocks.{block}.attn.to_out.alpha",
                    "transformer.single_transformer_blocks.{block}.attn.to_out.alpha",
                    "diffusion_model.single_transformer_blocks.{block}.attn.to_out.alpha",
                    "lora_unet_single_transformer_blocks_{block}_attn_to_out.alpha",
                ],
            ),
        ]

    @staticmethod
    def _get_bfl_double_stream_block_targets() -> list[LoRATarget]:
        return [
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_q",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_qkv.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_up.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_B.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_qkv.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_down.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_A.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_qkv.alpha",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.alpha",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.alpha",
                ],
                up_transform=LoraTransforms.split_q_up,
                down_transform=LoraTransforms.split_q_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_k",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_qkv.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_up.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_B.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_qkv.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_down.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_A.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_qkv.alpha",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.alpha",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.alpha",
                ],
                up_transform=LoraTransforms.split_k_up,
                down_transform=LoraTransforms.split_k_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_v",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_qkv.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_up.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_B.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_qkv.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.lora_down.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_A.weight",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_qkv.alpha",
                    "diffusion_model.double_blocks.{block}.img_attn.qkv.alpha",
                    "base_model.model.double_blocks.{block}.img_attn.qkv.alpha",
                ],
                up_transform=LoraTransforms.split_v_up,
                down_transform=LoraTransforms.split_v_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_out",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_proj.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.proj.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.proj.lora_up.weight",
                    "base_model.model.double_blocks.{block}.img_attn.proj.lora_B.weight",
                    "base_model.model.double_blocks.{block}.img_attn.proj.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_proj.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.proj.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.img_attn.proj.lora_down.weight",
                    "base_model.model.double_blocks.{block}.img_attn.proj.lora_A.weight",
                    "base_model.model.double_blocks.{block}.img_attn.proj.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_img_attn_proj.alpha",
                    "diffusion_model.double_blocks.{block}.img_attn.proj.alpha",
                    "base_model.model.double_blocks.{block}.img_attn.proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_q_proj",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_qkv.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_up.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_B.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_qkv.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_down.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_A.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_qkv.alpha",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.alpha",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.alpha",
                ],
                up_transform=LoraTransforms.split_q_up,
                down_transform=LoraTransforms.split_q_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_k_proj",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_qkv.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_up.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_B.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_qkv.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_down.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_A.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_qkv.alpha",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.alpha",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.alpha",
                ],
                up_transform=LoraTransforms.split_k_up,
                down_transform=LoraTransforms.split_k_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.add_v_proj",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_qkv.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_up.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_B.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_qkv.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.lora_down.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_A.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_qkv.alpha",
                    "diffusion_model.double_blocks.{block}.txt_attn.qkv.alpha",
                    "base_model.model.double_blocks.{block}.txt_attn.qkv.alpha",
                ],
                up_transform=LoraTransforms.split_v_up,
                down_transform=LoraTransforms.split_v_down,
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.attn.to_add_out",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_proj.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.proj.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.proj.lora_up.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.proj.lora_B.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.proj.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_proj.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.proj.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.txt_attn.proj.lora_down.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.proj.lora_A.weight",
                    "base_model.model.double_blocks.{block}.txt_attn.proj.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_txt_attn_proj.alpha",
                    "diffusion_model.double_blocks.{block}.txt_attn.proj.alpha",
                    "base_model.model.double_blocks.{block}.txt_attn.proj.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff.linear_in",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_img_mlp_0.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.img_mlp.0.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.img_mlp.0.lora_up.weight",
                    "base_model.model.double_blocks.{block}.img_mlp.0.lora_B.weight",
                    "base_model.model.double_blocks.{block}.img_mlp.0.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_img_mlp_0.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.img_mlp.0.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.img_mlp.0.lora_down.weight",
                    "base_model.model.double_blocks.{block}.img_mlp.0.lora_A.weight",
                    "base_model.model.double_blocks.{block}.img_mlp.0.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_img_mlp_0.alpha",
                    "diffusion_model.double_blocks.{block}.img_mlp.0.alpha",
                    "base_model.model.double_blocks.{block}.img_mlp.0.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff.linear_out",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_img_mlp_2.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.img_mlp.2.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.img_mlp.2.lora_up.weight",
                    "base_model.model.double_blocks.{block}.img_mlp.2.lora_B.weight",
                    "base_model.model.double_blocks.{block}.img_mlp.2.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_img_mlp_2.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.img_mlp.2.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.img_mlp.2.lora_down.weight",
                    "base_model.model.double_blocks.{block}.img_mlp.2.lora_A.weight",
                    "base_model.model.double_blocks.{block}.img_mlp.2.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_img_mlp_2.alpha",
                    "diffusion_model.double_blocks.{block}.img_mlp.2.alpha",
                    "base_model.model.double_blocks.{block}.img_mlp.2.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff_context.linear_in",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_txt_mlp_0.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.txt_mlp.0.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.txt_mlp.0.lora_up.weight",
                    "base_model.model.double_blocks.{block}.txt_mlp.0.lora_B.weight",
                    "base_model.model.double_blocks.{block}.txt_mlp.0.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_txt_mlp_0.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.txt_mlp.0.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.txt_mlp.0.lora_down.weight",
                    "base_model.model.double_blocks.{block}.txt_mlp.0.lora_A.weight",
                    "base_model.model.double_blocks.{block}.txt_mlp.0.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_txt_mlp_0.alpha",
                    "diffusion_model.double_blocks.{block}.txt_mlp.0.alpha",
                    "base_model.model.double_blocks.{block}.txt_mlp.0.alpha",
                ],
            ),
            LoRATarget(
                model_path="transformer_blocks.{block}.ff_context.linear_out",
                possible_up_patterns=[
                    "lora_unet_double_blocks_{block}_txt_mlp_2.lora_up.weight",
                    "diffusion_model.double_blocks.{block}.txt_mlp.2.lora_B.weight",
                    "diffusion_model.double_blocks.{block}.txt_mlp.2.lora_up.weight",
                    "base_model.model.double_blocks.{block}.txt_mlp.2.lora_B.weight",
                    "base_model.model.double_blocks.{block}.txt_mlp.2.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_double_blocks_{block}_txt_mlp_2.lora_down.weight",
                    "diffusion_model.double_blocks.{block}.txt_mlp.2.lora_A.weight",
                    "diffusion_model.double_blocks.{block}.txt_mlp.2.lora_down.weight",
                    "base_model.model.double_blocks.{block}.txt_mlp.2.lora_A.weight",
                    "base_model.model.double_blocks.{block}.txt_mlp.2.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_double_blocks_{block}_txt_mlp_2.alpha",
                    "diffusion_model.double_blocks.{block}.txt_mlp.2.alpha",
                    "base_model.model.double_blocks.{block}.txt_mlp.2.alpha",
                ],
            ),
        ]

    @staticmethod
    def _get_bfl_single_stream_block_targets() -> list[LoRATarget]:
        # Flux2 single blocks fuse qkv+mlp into one projection, matching BFL "linear1".
        return [
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_qkv_mlp_proj",
                possible_up_patterns=[
                    "lora_unet_single_blocks_{block}_linear1.lora_up.weight",
                    "diffusion_model.single_blocks.{block}.linear1.lora_B.weight",
                    "diffusion_model.single_blocks.{block}.linear1.lora_up.weight",
                    "base_model.model.single_blocks.{block}.linear1.lora_B.weight",
                    "base_model.model.single_blocks.{block}.linear1.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_single_blocks_{block}_linear1.lora_down.weight",
                    "diffusion_model.single_blocks.{block}.linear1.lora_A.weight",
                    "diffusion_model.single_blocks.{block}.linear1.lora_down.weight",
                    "base_model.model.single_blocks.{block}.linear1.lora_A.weight",
                    "base_model.model.single_blocks.{block}.linear1.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_single_blocks_{block}_linear1.alpha",
                    "diffusion_model.single_blocks.{block}.linear1.alpha",
                    "base_model.model.single_blocks.{block}.linear1.alpha",
                ],
            ),
            LoRATarget(
                model_path="single_transformer_blocks.{block}.attn.to_out",
                possible_up_patterns=[
                    "lora_unet_single_blocks_{block}_linear2.lora_up.weight",
                    "diffusion_model.single_blocks.{block}.linear2.lora_B.weight",
                    "diffusion_model.single_blocks.{block}.linear2.lora_up.weight",
                    "base_model.model.single_blocks.{block}.linear2.lora_B.weight",
                    "base_model.model.single_blocks.{block}.linear2.lora_up.weight",
                ],
                possible_down_patterns=[
                    "lora_unet_single_blocks_{block}_linear2.lora_down.weight",
                    "diffusion_model.single_blocks.{block}.linear2.lora_A.weight",
                    "diffusion_model.single_blocks.{block}.linear2.lora_down.weight",
                    "base_model.model.single_blocks.{block}.linear2.lora_A.weight",
                    "base_model.model.single_blocks.{block}.linear2.lora_down.weight",
                ],
                possible_alpha_patterns=[
                    "lora_unet_single_blocks_{block}_linear2.alpha",
                    "diffusion_model.single_blocks.{block}.linear2.alpha",
                    "base_model.model.single_blocks.{block}.linear2.alpha",
                ],
            ),
        ]



import mlx.core as mx

from mflux.models.common.tokenizer import Tokenizer
from mflux.models.flux2.model.flux2_text_encoder.qwen3_text_encoder import Qwen3TextEncoder


class Flux2PromptEncoder:
    @staticmethod
    def encode_prompt(
        prompt: str | list[str],
        tokenizer: Tokenizer,
        text_encoder: Qwen3TextEncoder,
        num_images_per_prompt: int = 1,
        max_sequence_length: int = 512,
        text_encoder_out_layers: tuple[int, ...] = (9, 18, 27),
    ) -> tuple[mx.array, mx.array]:
        prompt_embeds = Flux2PromptEncoder._get_qwen3_prompt_embeds(
            prompt=prompt,
            tokenizer=tokenizer,
            text_encoder=text_encoder,
            max_sequence_length=max_sequence_length,
            hidden_state_layers=text_encoder_out_layers,
        )
        if num_images_per_prompt > 1:
            prompt_embeds = mx.repeat(prompt_embeds, num_images_per_prompt, axis=0)
        text_ids = Flux2PromptEncoder.prepare_text_ids(prompt_embeds)
        return prompt_embeds, text_ids

    @staticmethod
    def _get_qwen3_prompt_embeds(
        prompt: str | list[str],
        tokenizer: Tokenizer,
        text_encoder: Qwen3TextEncoder,
        max_sequence_length: int,
        hidden_state_layers: tuple[int, ...],
    ) -> mx.array:
        tokens = tokenizer.tokenize(prompt=prompt, max_length=max_sequence_length)
        return text_encoder.get_prompt_embeds(
            input_ids=tokens.input_ids,
            attention_mask=tokens.attention_mask,
            hidden_state_layers=hidden_state_layers,
        )

    @staticmethod
    def prepare_text_ids(x: mx.array, t_coord: mx.array | None = None) -> mx.array:
        batch_size, seq_len, _ = x.shape
        out_ids = []
        for i in range(batch_size):
            if t_coord is None:
                t = mx.zeros((seq_len,), dtype=mx.int32)
            else:
                t = t_coord[i]
                if t.ndim == 0:
                    t = mx.full((seq_len,), t, dtype=mx.int32)
                elif t.shape[0] != seq_len:
                    t = mx.broadcast_to(t, (seq_len,))
                t = t.astype(mx.int32)
            h = mx.zeros((seq_len,), dtype=mx.int32)
            w = mx.zeros((seq_len,), dtype=mx.int32)
            token_ids = mx.arange(seq_len, dtype=mx.int32)
            coords = mx.stack([t, h, w, token_ids], axis=1)
            out_ids.append(coords)
        return mx.stack(out_ids, axis=0)



from __future__ import annotations

from pathlib import Path
from typing import Any

import mlx.core as mx

from mflux.models.common.config.config import Config
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.flux2.training_adapter.flux2_base_training_adapter import Flux2BaseTrainingAdapter
from mflux.models.flux2.variants.txt2img.flux2_klein import Flux2Klein


class Flux2TrainingAdapter(Flux2BaseTrainingAdapter):
    def __init__(self, *, model_config: ModelConfig, quantize: int | None, model_path: str | None = None):
        super().__init__(
            model_config=model_config,
            quantize=quantize,
            model_factory=lambda cfg, q: Flux2Klein(model_config=cfg, quantize=q, model_path=model_path),
        )

    def encode_data(
        self,
        *,
        data_id: int,
        image_path: Path,
        prompt: str,
        width: int,
        height: int,
        input_image_path: Path | None = None,
    ) -> tuple[mx.array, Any]:
        clean_latents, img_ids = self._encode_output_latents(image_path=image_path, width=width, height=height)
        prompt_embeds, text_ids = self._encode_prompt(prompt=prompt)

        mx.eval(clean_latents, prompt_embeds, text_ids, img_ids)
        return clean_latents, {"prompt_embeds": prompt_embeds, "text_ids": text_ids, "img_ids": img_ids}

    def predict_noise(self, *, t: int, latents_t: mx.array, sigmas: mx.array, cond: Any, config: Config) -> mx.array:  # noqa: ARG002
        return self._flux2.transformer(
            hidden_states=latents_t,
            encoder_hidden_states=cond["prompt_embeds"],
            timestep=config.scheduler.timesteps[t],
            img_ids=cond["img_ids"],
            txt_ids=cond["text_ids"],
            guidance=None,
        )

    def generate_preview_image(
        self,
        *,
        seed: int,
        prompt: str,
        width: int,
        height: int,
        steps: int,
        image_paths: list[Path | str] | None = None,
    ):
        with self._assistant_disabled():
            image = self._flux2.generate_image(
                seed=seed,
                prompt=prompt,
                num_inference_steps=steps,
                height=height,
                width=width,
                guidance=self._guidance,
            )
        self._flux2.prompt_cache = {}
        return image.image



import mlx.core as mx

from mflux.models.common.config.model_config import ModelConfig


class Flux2LatentCreator:
    @staticmethod
    def patchify_latents(latents: mx.array) -> mx.array:
        if latents.ndim == 5 and latents.shape[2] == 1:
            latents = latents[:, :, 0, :, :]
        if latents.ndim != 4:
            raise ValueError(f"Expected latents with ndim=4, got shape={latents.shape}")
        batch_size, num_channels, height, width = latents.shape
        latents = latents.reshape(batch_size, num_channels, height // 2, 2, width // 2, 2)
        latents = latents.transpose(0, 1, 3, 5, 2, 4)
        latents = latents.reshape(batch_size, num_channels * 4, height // 2, width // 2)
        return latents

    @staticmethod
    def pack_latents(latents: mx.array) -> mx.array:
        batch_size, num_channels, height, width = latents.shape
        return latents.reshape(batch_size, num_channels, height * width).transpose(0, 2, 1)

    @staticmethod
    def unpack_latents(latents: mx.array, height: int, width: int, vae_scale_factor: int = 8) -> mx.array:
        if latents.ndim == 4:
            return latents

        if latents.ndim != 3:
            raise ValueError(f"Expected packed latents with ndim=3, got shape={latents.shape}")

        batch_size, seq_len, channels = latents.shape
        latent_height = height // (vae_scale_factor * 2)
        latent_width = width // (vae_scale_factor * 2)
        expected_seq_len = latent_height * latent_width
        if expected_seq_len != seq_len:
            raise ValueError(
                f"Packed latent seq_len mismatch: got {seq_len}, expected {expected_seq_len} "
                f"for height={height} width={width} (latent {latent_height}x{latent_width})"
            )

        return latents.reshape(batch_size, latent_height, latent_width, channels).transpose(0, 3, 1, 2)

    @staticmethod
    def prepare_grid_ids(latents: mx.array, *, t_coord: int) -> mx.array:
        batch_size, _, height, width = latents.shape
        h_ids = mx.arange(height, dtype=mx.int32)
        w_ids = mx.arange(width, dtype=mx.int32)
        h_grid = mx.broadcast_to(mx.expand_dims(h_ids, axis=1), (height, width))
        w_grid = mx.broadcast_to(mx.expand_dims(w_ids, axis=0), (height, width))
        flat_h = h_grid.reshape(-1)
        flat_w = w_grid.reshape(-1)
        t = mx.full(flat_h.shape, t_coord, dtype=mx.int32)
        layer_ids = mx.zeros_like(flat_h)
        coords = mx.stack([t, flat_h, flat_w, layer_ids], axis=1)
        coords = mx.expand_dims(coords, axis=0)
        return mx.broadcast_to(coords, (batch_size, coords.shape[1], coords.shape[2]))

    @staticmethod
    def prepare_latents(
        seed: int,
        height: int,
        width: int,
        batch_size: int,
        num_latents_channels: int = 32,
        vae_scale_factor: int = 8,
    ) -> tuple[mx.array, mx.array, int, int]:
        height = 2 * (height // (vae_scale_factor * 2))
        width = 2 * (width // (vae_scale_factor * 2))
        latent_height = height // 2
        latent_width = width // 2
        latents = mx.random.normal(
            shape=(batch_size, num_latents_channels * 4, latent_height, latent_width),
            key=mx.random.key(seed),
        ).astype(ModelConfig.precision)
        latent_ids = Flux2LatentCreator.prepare_grid_ids(latents, t_coord=0)
        return latents, latent_ids, latent_height, latent_width

    @staticmethod
    def prepare_packed_latents(
        seed: int,
        height: int,
        width: int,
        batch_size: int,
        num_latents_channels: int = 32,
        vae_scale_factor: int = 8,
    ) -> tuple[mx.array, mx.array, int, int]:
        latents, latent_ids, latent_height, latent_width = Flux2LatentCreator.prepare_latents(
            seed=seed,
            height=height,
            width=width,
            batch_size=batch_size,
            num_latents_channels=num_latents_channels,
            vae_scale_factor=vae_scale_factor,
        )
        return Flux2LatentCreator.pack_latents(latents), latent_ids, latent_height, latent_width



import mlx.core as mx
from mlx import nn

from mflux.models.flux2.model.flux2_vae.common.downsample_2d import Flux2Downsample2D
from mflux.models.flux2.model.flux2_vae.common.resnet_block_2d import Flux2ResnetBlock2D


class Flux2DownEncoderBlock2D(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_layers: int = 2,
        eps: float = 1e-6,
        groups: int = 32,
        add_downsample: bool = True,
        downsample_padding: int = 0,
    ):
        super().__init__()
        self.resnets = [
            Flux2ResnetBlock2D(
                in_channels=in_channels if i == 0 else out_channels,
                out_channels=out_channels,
                eps=eps,
                groups=groups,
            )
            for i in range(num_layers)
        ]
        self.downsamplers = [Flux2Downsample2D(out_channels, padding=downsample_padding)] if add_downsample else []

    def __call__(self, hidden_states: mx.array) -> mx.array:
        for resnet in self.resnets:
            hidden_states = resnet(hidden_states)
        for downsampler in self.downsamplers:
            hidden_states = downsampler(hidden_states)
        return hidden_states



import json
from pathlib import Path

from mflux.models.common.training.state.training_spec import TrainingSpec
from mflux.models.common.training.state.training_state import (
    TRAINING_FILE_NAME_PREVIEW_IMAGE,
    TRAINING_PATH_PREVIEW_IMAGES,
    TrainingState,
)


class Plotter:
    @staticmethod
    def update_loss_plot(training_spec: TrainingSpec, training_state: TrainingState, target_loss: float = 0.3) -> None:
        stats = training_state.statistics
        if not stats.steps or not stats.losses:
            return

        total_step = int(training_state.iterator.total_number_of_steps())
        smooth_loss = False
        smooth_window = 5
        preview_suffix = "01"
        if training_spec.monitoring is not None:
            smooth_loss = training_spec.monitoring.smooth_loss
            smooth_window = training_spec.monitoring.smooth_loss_window
            if training_spec.monitoring.preview_prompt_names:
                preview_suffix = training_spec.monitoring.preview_prompt_names[0]

        steps = [int(step) for step in stats.steps]
        losses = [float(loss) for loss in stats.losses]
        max_x = max(steps)
        initial_padding = 0.4
        final_padding = 0.01
        padding_limit = initial_padding - (initial_padding - final_padding) * (max_x / total_step)
        padding = max_x * padding_limit

        preview_rel_root = Path("..") / TRAINING_PATH_PREVIEW_IMAGES
        preview_abs_root = Path(training_spec.checkpoint.output_path) / TRAINING_PATH_PREVIEW_IMAGES
        preview_step_to_src: dict[int, str] = {}
        for step in steps:
            preview_name = f"{int(step):07d}_{TRAINING_FILE_NAME_PREVIEW_IMAGE}_{preview_suffix}.png"
            preview_src = (preview_rel_root / preview_name).as_posix()
            preview_abs = preview_abs_root / preview_name
            if preview_abs.exists():
                preview_step_to_src[int(step)] = preview_src

        if 0 not in preview_step_to_src:
            preview_zero = preview_abs_root / f"{0:07d}_{TRAINING_FILE_NAME_PREVIEW_IMAGE}_{preview_suffix}.png"
            if preview_zero.exists():
                preview_step_to_src[0] = (preview_rel_root / preview_zero.name).as_posix()

        preview_steps = sorted(preview_step_to_src.keys())
        loss_by_step = {int(step): float(loss) for step, loss in zip(steps, losses)}
        smoothed = Plotter._smooth_losses(losses, smooth_window)

        data = {
            "steps": steps,
            "losses": losses,
            "smoothed": smoothed,
            "smoothEnabled": smooth_loss,
            "previewMap": preview_step_to_src,
            "previewSteps": preview_steps,
            "lossByStep": loss_by_step,
            "padding": padding,
            "totalStep": total_step,
        }
        html = Plotter._build_html(data)
        path = training_state.get_current_loss_plot_path(training_spec)
        with open(path, "w", encoding="utf-8") as f:
            f.write(html)

    @staticmethod
    def _smooth_losses(values: list[float], window: int) -> list[float]:
        if window <= 1:
            return list(values)
        smoothed: list[float] = []
        for i, val in enumerate(values):
            start = max(0, i - window + 1)
            window_values = values[start : i + 1]
            smoothed.append(float(sum(window_values)) / float(len(window_values)))
        return smoothed

    @staticmethod
    def _build_html(data: dict) -> str:
        json_data = json.dumps(data)
        return f"""<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <title>Loss</title>
    <style>
      :root {{
        --bg: #F4EFE6;
        --panel: #F8F4EC;
        --text: #2F2A24;
        --muted: #6B6258;
        --grid: rgba(148, 163, 184, 0.18);
        --axis: rgba(148, 163, 184, 0.28);
        --line: #4A6FA5;
        --point: #3C5B87;
        --preview: #5B8BD9;
      }}
      * {{ box-sizing: border-box; }}
      body {{
        margin: 0;
        background: var(--bg);
        font-family: Inter, ui-sans-serif, system-ui, -apple-system, Segoe UI, sans-serif;
        color: var(--text);
        min-height: 100vh;
        display: flex;
        justify-content: center;
        align-items: center;
      }}
      .container {{
        max-width: 1400px;
        margin: 8px auto;
        padding: 0 20px 12px;
        width: 100%;
      }}
      .card {{
        background: var(--panel);
        border-radius: 18px;
        border: 1px solid rgba(198, 184, 166, 0.5);
        box-shadow: 0 18px 40px rgba(55, 48, 40, 0.12);
        padding: 12px 12px 14px;
        position: relative;
      }}
      .controls {{
        position: absolute;
        top: 14px;
        right: 16px;
        display: flex;
        flex-direction: column;
        align-items: flex-end;
        gap: 10px;
        font-size: 12px;
        color: var(--muted);
        z-index: 2;
      }}
      .control-row {{
        display: flex;
        align-items: center;
        gap: 10px;
      }}
      .toggle {{
        position: relative;
        display: inline-block;
        width: 44px;
        height: 24px;
      }}
      .toggle input {{
        opacity: 0;
        width: 0;
        height: 0;
      }}
      .toggle-slider {{
        position: absolute;
        inset: 0;
        background: rgba(148, 163, 184, 0.35);
        border-radius: 999px;
        transition: background 160ms ease;
      }}
      .toggle-slider::before {{
        content: '';
        position: absolute;
        width: 20px;
        height: 20px;
        left: 2px;
        top: 2px;
        background: #fffaf2;
        border-radius: 50%;
        box-shadow: 0 2px 6px rgba(55, 48, 40, 0.2);
        transition: transform 160ms ease;
      }}
      .toggle input:checked + .toggle-slider {{
        background: #7BA4E8;
      }}
      .toggle.is-blue input:checked + .toggle-slider {{
        background: #4A6FA5;
      }}
      .toggle.is-red input:checked + .toggle-slider {{
        background: #D14B4B;
      }}
      .toggle input:checked + .toggle-slider::before {{
        transform: translateX(20px);
      }}
      .chart-wrap {{
        position: relative;
      }}
      #chart {{
        width: 100%;
        height: 86vh;
        display: block;
      }}
      .preview-popovers {{
        position: absolute;
        inset: 0;
        pointer-events: none;
      }}
      .popover-card {{
        position: absolute;
        --popover-bg: rgba(255, 250, 242, 0.98);
        padding: 12px 14px;
        background: var(--popover-bg);
        border: 1px solid rgba(198, 184, 166, 0.7);
        border-radius: 14px;
        box-shadow: 0 18px 40px rgba(55, 48, 40, 0.18);
        font-size: 12px;
        color: #3F3A33;
      }}
      .popover-card .pointer {{
        position: absolute;
        bottom: -9px;
        left: 50%;
        width: 16px;
        height: 10px;
        transform: translateX(-50%);
      }}
      .popover-card .pointer::before {{
        content: '';
        position: absolute;
        inset: 0;
        background: rgba(198, 184, 166, 0.7);
        clip-path: polygon(50% 100%, 0 0, 100% 0);
      }}
      .popover-card .pointer::after {{
        content: '';
        position: absolute;
        left: 1px;
        right: 1px;
        top: 0;
        bottom: -1px;
        background: var(--popover-bg);
        clip-path: polygon(50% 100%, 0 0, 100% 0);
      }}
      .popover-card .step {{
        font-weight: 600;
        margin-bottom: 4px;
        color: #2F2A24;
      }}
      .popover-card .loss {{
        color: #5B5247;
        margin-bottom: 8px;
      }}
      .popover-card img {{
        width: 130px;
        height: auto;
        border-radius: 12px;
        display: block;
        box-shadow: 0 4px 10px rgba(55, 48, 40, 0.12);
      }}

      .preview-popover {{
        position: absolute;
        --scale: 0.82;
        transform: translate(-50%, -100%) scale(var(--scale));
        transform-origin: center bottom;
        transition: transform 140ms ease, box-shadow 140ms ease;
        pointer-events: auto;
      }}
      .preview-popover.is-active {{
        --scale: 1.6;
        box-shadow: 0 20px 40px rgba(55, 48, 40, 0.2);
      }}
      .preview-popover:hover {{
        --scale: 1.6;
        box-shadow: 0 20px 40px rgba(55, 48, 40, 0.2);
      }}
      .popover {{
        position: fixed;
        pointer-events: none;
        display: none;
        z-index: 9999;
      }}
    </style>
  </head>
  <body>
    <div class="container">
      <div class="card">
        <div class="controls">
          <div class="control-row">
            <span>Loss</span>
            <label class="toggle is-blue">
              <input type="checkbox" id="toggle-loss" checked />
              <span class="toggle-slider"></span>
            </label>
          </div>
          <div class="control-row">
            <span>Previews</span>
            <label class="toggle">
              <input type="checkbox" id="toggle-previews" checked />
              <span class="toggle-slider"></span>
            </label>
          </div>
          <div class="control-row" id="smooth-control">
            <span>Smooth</span>
            <label class="toggle is-red">
              <input type="checkbox" id="toggle-smooth" />
              <span class="toggle-slider"></span>
            </label>
          </div>
        </div>
        <div class="chart-wrap" id="chart-wrap">
          <canvas id="chart"></canvas>
          <div class="preview-popovers" id="preview-popovers"></div>
        </div>
      </div>
    </div>
    <div class="popover popover-card" id="popover"></div>
    <script>
      const data = {json_data};
      const chartWrap = document.getElementById('chart-wrap');
      const canvas = document.getElementById('chart');
      const ctx = canvas.getContext('2d');
      const previewLayer = document.getElementById('preview-popovers');
      const lossToggle = document.getElementById('toggle-loss');
      const previewToggle = document.getElementById('toggle-previews');
      const smoothToggle = document.getElementById('toggle-smooth');
      const smoothControl = document.getElementById('smooth-control');
      const popover = document.getElementById('popover');
      function buildPopoverHTML(step, loss, src) {{
        const img = src ? '<img src="' + src + '" alt="" />' : '';
        return (
          '<div class="pointer"></div>' +
          '<div class="step">step: ' + step + '</div>' +
          '<div class="loss">loss: ' + loss.toFixed(4) + '</div>' +
          img
        );
      }}
      const dpr = window.devicePixelRatio || 1;
      const steps = data.steps;
      const losses = data.losses;
      const smoothed = data.smoothed || [];
      const smoothEnabled = data.smoothEnabled || false;
      let showSmoothed = smoothEnabled;
      let showLoss = true;
      const previewMapAll = data.previewMap || {{}};
      const lossByStep = data.lossByStep || {{}};
      const previewSteps = data.previewSteps || [];
      const maxPreviewPopovers = 8;

      const margin = {{ left: 72, right: 30, top: 120, bottom: 60 }};

      function updateLossVisibility() {{
        if (!lossToggle) return;
        showLoss = lossToggle.checked;
        draw();
      }}

      function updateSmoothVisibility() {{
        if (!smoothToggle) return;
        showSmoothed = smoothToggle.checked;
        draw();
      }}

      function updatePreviewVisibility() {{
        if (!previewLayer || !previewToggle) return;
        const enabled = previewToggle.checked;
        previewLayer.style.display = enabled ? 'block' : 'none';
        if (!enabled) setActivePreview(null);
      }}

      function resize() {{
        const width = canvas.clientWidth || canvas.parentElement.clientWidth;
        const height = canvas.clientHeight || 520;
        canvas.style.width = width + 'px';
        canvas.style.height = height + 'px';
        canvas.width = width * dpr;
        canvas.height = height * dpr;
        ctx.setTransform(1, 0, 0, 1, 0, 0);
        ctx.scale(dpr, dpr);
        draw();
      }}

      function scaleX(step, minStep, maxStep, width) {{
        if (maxStep === minStep) return margin.left + width / 2;
        return margin.left + ((step - minStep) / (maxStep - minStep)) * width;
      }}

      function scaleY(value, minY, maxY, height) {{
        if (maxY === minY) return margin.top + height / 2;
        return margin.top + (1 - (value - minY) / (maxY - minY)) * height;
      }}

      function niceStep(range, targetTicks, integerOnly) {{
        if (range <= 0) return 1;
        const rough = range / Math.max(1, targetTicks);
        const pow10 = Math.pow(10, Math.floor(Math.log10(rough)));
        const candidates = [1, 2, 5, 10].map((n) => n * pow10);
        let step = candidates.find((n) => n >= rough) || candidates[candidates.length - 1];
        if (integerOnly) {{
          step = Math.max(1, Math.round(step));
        }}
        return step;
      }}

      function drawGrid(width, height, minStep, maxStep, minY, maxY) {{
        ctx.strokeStyle = getComputedStyle(document.documentElement).getPropertyValue('--grid').trim();
        ctx.lineWidth = 1;
        const xTicks = 7;
        const yTicks = 5;
        const xRange = Math.max(1, maxStep - minStep);
        const yRange = Math.max(0.0001, maxY - minY);
        const xStep = niceStep(xRange, xTicks, true);
        const yStep = niceStep(yRange, yTicks, false);

        const xStart = Math.ceil(minStep / xStep) * xStep;
        const xEnd = Math.floor(maxStep / xStep) * xStep;
        for (let xVal = xStart; xVal <= xEnd; xVal += xStep) {{
          const x = scaleX(xVal, minStep, maxStep, width);
          ctx.beginPath();
          ctx.moveTo(x, margin.top);
          ctx.lineTo(x, margin.top + height);
          ctx.stroke();
        }}
        const yStart = Math.ceil(minY / yStep) * yStep;
        const yEnd = Math.floor(maxY / yStep) * yStep;
        for (let yVal = yStart; yVal <= yEnd + (yStep * 0.5); yVal += yStep) {{
          const y = scaleY(yVal, minY, maxY, height);
          ctx.beginPath();
          ctx.moveTo(margin.left, y);
          ctx.lineTo(margin.left + width, y);
          ctx.stroke();
        }}
        ctx.strokeStyle = getComputedStyle(document.documentElement).getPropertyValue('--axis').trim();
        ctx.beginPath();
        ctx.moveTo(margin.left, margin.top + height);
        ctx.lineTo(margin.left + width, margin.top + height);
        ctx.stroke();
        ctx.beginPath();
        ctx.moveTo(margin.left, margin.top);
        ctx.lineTo(margin.left, margin.top + height);
        ctx.stroke();
        ctx.fillStyle = getComputedStyle(document.documentElement).getPropertyValue('--muted').trim();
        ctx.textAlign = 'center';
        ctx.textBaseline = 'top';
        ctx.font = '600 14px Inter, ui-sans-serif, system-ui';
        ctx.fillText('Steps', margin.left + width / 2, margin.top + height + 48);
        for (let xVal = xStart; xVal <= xEnd; xVal += xStep) {{
          const x = scaleX(xVal, minStep, maxStep, width);
          ctx.font = '12px Inter, ui-sans-serif, system-ui';
          ctx.fillText(String(xVal), x, margin.top + height + 18);
        }}
        ctx.textAlign = 'right';
        ctx.textBaseline = 'middle';
        for (let yVal = yStart; yVal <= yEnd + (yStep * 0.5); yVal += yStep) {{
          const y = scaleY(yVal, minY, maxY, height);
          ctx.fillText(yVal.toFixed(3), margin.left - 12, y);
        }}
        ctx.save();
        ctx.translate(6, margin.top + height / 2 + 22);
        ctx.rotate(-Math.PI / 2);
        ctx.font = '600 14px Inter, ui-sans-serif, system-ui';
        ctx.fillText('Loss', 0, 0);
        ctx.restore();
      }}

      function linePath(points) {{
        ctx.beginPath();
        points.forEach((p, i) => {{
          if (i === 0) ctx.moveTo(p.x, p.y);
          else ctx.lineTo(p.x, p.y);
        }});
        ctx.stroke();
      }}

      function draw() {{
        if (!steps.length) return;
        const width = canvas.width / dpr - margin.left - margin.right;
        const height = canvas.height / dpr - margin.top - margin.bottom;
        ctx.clearRect(0, 0, canvas.width, canvas.height);
        ctx.fillStyle = getComputedStyle(document.documentElement).getPropertyValue('--panel').trim();
        ctx.fillRect(0, 0, canvas.width / dpr, canvas.height / dpr);

        const minStep = Math.min(...steps);
        const maxStep = Math.max(...steps) + data.padding;
        const allY = smoothed.length ? losses.concat(smoothed) : losses;
        let minY = Math.min(...allY);
        let maxY = Math.max(...allY);
        if (minY === maxY) {{
          minY -= 0.01;
          maxY += 0.01;
        }} else {{
          const pad = (maxY - minY) * 0.06;
          minY -= pad;
          maxY += pad;
        }}

        drawGrid(width, height, minStep, maxStep, minY, maxY);

        const previewMap = previewMapAll;

        const points = steps.map((step, i) => {{
          const previewSrc = previewMap[String(step)] || previewMap[step] || '';
          return {{
            step,
            loss: losses[i],
            x: scaleX(step, minStep, maxStep, width),
            y: scaleY(losses[i], minY, maxY, height),
            preview: previewSrc
          }};
        }});

        if (showLoss) {{
          ctx.strokeStyle = getComputedStyle(document.documentElement).getPropertyValue('--line').trim();
          ctx.lineWidth = 2.5;
          linePath(points);
        }}

        if (showSmoothed && smoothed.length) {{
          const smoothPoints = steps.map((step, i) => {{
            return {{
              x: scaleX(step, minStep, maxStep, width),
              y: scaleY(smoothed[i], minY, maxY, height),
            }};
          }});
          ctx.setLineDash([6, 6]);
          ctx.strokeStyle = '#D14B4B';
          ctx.lineWidth = 2;
          linePath(smoothPoints);
          ctx.setLineDash([]);
        }}

        if (showLoss) {{
          points.forEach((p) => {{
            ctx.beginPath();
            ctx.fillStyle = p.preview ? getComputedStyle(document.documentElement).getPropertyValue('--preview').trim() : getComputedStyle(document.documentElement).getPropertyValue('--point').trim();
            ctx.arc(p.x, p.y, p.preview ? 4.5 : 3.5, 0, Math.PI * 2);
            ctx.fill();
          }});

          Object.keys(previewMap).forEach((stepKey) => {{
            const step = Number(stepKey);
            const loss = lossByStep[step] ?? lossByStep[String(step)] ?? losses[0];
            const x = scaleX(step, minStep, maxStep, width);
            const y = scaleY(loss, minY, maxY, height);
            ctx.beginPath();
            ctx.fillStyle = getComputedStyle(document.documentElement).getPropertyValue('--preview').trim();
            ctx.arc(x, y, 4.5, 0, Math.PI * 2);
            ctx.fill();
          }});
        }}

        window.__lossPoints = points.map((p) => {{
          const previewSrc = previewMap[String(p.step)] || previewMap[p.step] || '';
          return {{
            step: p.step,
            loss: p.loss,
            x: p.x,
            y: p.y,
            preview: previewSrc
          }};
        }});
        renderPreviewPopovers(points, previewMap);
        updatePreviewVisibility();
      }}

      function selectPreviewSteps(steps, maxCount) {{
        if (!steps || !steps.length) return [];
        if (steps.length <= maxCount) return steps.slice();
        if (maxCount <= 1) return [steps[0]];
        const selected = [];
        const lastIndex = steps.length - 1;
        for (let i = 0; i < maxCount; i++) {{
          const idx = Math.round((i * lastIndex) / (maxCount - 1));
          selected.push(steps[idx]);
        }}
        return Array.from(new Set(selected));
      }}

      function renderPreviewPopovers(points, previewMap) {{
        if (!previewLayer || !chartWrap) return;
        previewLayer.innerHTML = '';
        const allowedSteps = new Set(selectPreviewSteps(previewSteps, maxPreviewPopovers).map(String));
        const wrapRect = chartWrap.getBoundingClientRect();
        const padding = 8;
        const overflowPadding = 24;
        const baseScale = 0.82;
        points.forEach((p) => {{
          const src = previewMap[String(p.step)] || previewMap[p.step];
          if (allowedSteps.size && !allowedSteps.has(String(p.step))) return;
          if (!src) return;
          const pop = document.createElement('div');
          pop.className = 'preview-popover popover-card';
          pop.dataset.step = String(p.step);
          pop.innerHTML = buildPopoverHTML(p.step, p.loss, src);
          previewLayer.appendChild(pop);
          const popRect = pop.getBoundingClientRect();
          const width = popRect.width / baseScale;
          const height = popRect.height / baseScale;
          const minCenter = -overflowPadding + width / 2;
          const maxCenter = Math.max(minCenter, wrapRect.width - padding - width / 2);
          const centerX = Math.min(Math.max(p.x, minCenter), maxCenter);
          const minAnchor = padding + height;
          const maxAnchor = wrapRect.height - padding;
          const anchorY = Math.min(Math.max(p.y - 18, minAnchor), maxAnchor);
          pop.style.left = centerX + 'px';
          pop.style.top = anchorY + 'px';
          const pointer = pop.querySelector('.pointer');
          if (pointer) {{
            const pointerLeft = Math.min(Math.max(p.x - (centerX - width / 2), 10), width - 10);
            pointer.style.left = pointerLeft + 'px';
          }}
        }});
      }}

      function setActivePreview(step) {{
        if (!previewLayer) return;
        const activeStep = step === null || step === undefined ? null : String(step);
        previewLayer.querySelectorAll('.preview-popover').forEach((pop) => {{
          if (activeStep && pop.dataset.step === activeStep) {{
            pop.classList.add('is-active');
          }} else {{
            pop.classList.remove('is-active');
          }}
        }});
      }}

      function findClosestPoint(x, y) {{
        const points = window.__lossPoints || [];
        let best = null;
        let bestDist = Infinity;
        points.forEach((p) => {{
          const dx = p.x - x;
          const dy = p.y - y;
          const dist = Math.sqrt(dx * dx + dy * dy);
          if (dist < bestDist) {{
            bestDist = dist;
            best = p;
          }}
        }});
        if (best && bestDist <= 18) return best;
        return null;
      }}

      function showPopover(point, clientX, clientY) {{
        if (!point) {{
          popover.style.display = 'none';
          return;
        }}
        const previewSrc = point.preview || previewMapAll[String(point.step)] || previewMapAll[point.step] || '';
        if (previewSrc) {{
          popover.style.display = 'none';
          setActivePreview(point.step);
          return;
        }}
        setActivePreview(null);
        popover.innerHTML = buildPopoverHTML(point.step, point.loss, '');
        popover.style.display = 'block';
        const popoverWidth = popover.offsetWidth;
        const popoverHeight = popover.offsetHeight;
        const viewportWidth = window.innerWidth || document.documentElement.clientWidth;
        const desiredLeft = clientX - (popoverWidth / 2);
        const minLeft = 12;
        const maxLeft = Math.max(minLeft, viewportWidth - popoverWidth - 12);
        const clampedLeft = Math.min(Math.max(desiredLeft, minLeft), maxLeft);
        popover.style.left = clampedLeft + 'px';
        const viewportHeight = window.innerHeight || document.documentElement.clientHeight;
        const desiredTop = clientY - popoverHeight - 18;
        const minTop = 12;
        const maxTop = Math.max(minTop, viewportHeight - popoverHeight - 12);
        const clampedTop = Math.min(Math.max(desiredTop, minTop), maxTop);
        popover.style.top = clampedTop + 'px';
        const pointer = popover.querySelector('.pointer');
        if (pointer) {{
          const pointerCenter = clientX - clampedLeft;
          const pointerMin = 16;
          const pointerMax = popoverWidth - 16;
          const pointerLeft = Math.min(Math.max(pointerCenter, pointerMin), pointerMax);
          pointer.style.left = pointerLeft + 'px';
        }}
      }}

      canvas.addEventListener('mousemove', (e) => {{
        const rect = canvas.getBoundingClientRect();
        const x = e.clientX - rect.left;
        const y = e.clientY - rect.top;
        const point = findClosestPoint(x, y);
        showPopover(point, e.clientX, e.clientY);
      }});
      canvas.addEventListener('mouseleave', () => {{
        popover.style.display = 'none';
        setActivePreview(null);
      }});

      if (lossToggle) {{
        lossToggle.checked = showLoss;
        lossToggle.addEventListener('change', updateLossVisibility);
      }}
      if (previewToggle) {{
        previewToggle.addEventListener('change', updatePreviewVisibility);
      }}
      if (smoothControl) {{
        smoothControl.style.display = smoothed.length ? 'flex' : 'none';
      }}
      if (smoothToggle) {{
        smoothToggle.checked = showSmoothed;
        smoothToggle.addEventListener('change', updateSmoothVisibility);
      }}
      window.addEventListener('resize', resize);
      resize();
    </script>
  </body>
</html>
"""



from pathlib import Path

from mflux.callbacks.callback_manager import CallbackManager
from mflux.cli.defaults import defaults as ui_defaults
from mflux.cli.parser.parsers import CommandLineParser
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.fibo.latent_creator.fibo_latent_creator import FiboLatentCreator
from mflux.models.fibo.variants.edit.fibo_edit import FIBOEdit
from mflux.models.fibo.variants.edit.util import FIBO_EDIT_RMBG_DEFAULT_JSON_PROMPT, FiboEditUtil
from mflux.utils.dimension_resolver import DimensionResolver
from mflux.utils.exceptions import ModelConfigError, PromptFileReadError, StopImageGenerationException
from mflux.utils.generated_image import GeneratedImage
from mflux.utils.prompt_util import PromptUtil

FIBO_EDIT_RMBG_GUIDANCE_DEFAULT = 1.0


def _resolve_fibo_edit_model_config(parser: CommandLineParser, args) -> ModelConfig:
    try:
        model_config = ModelConfig.from_name(model_name=args.model, base_model=args.base_model)
    except ModelConfigError as exc:
        parser.error(str(exc))

    compatible_edit_aliases = {"fibo-edit", "fibo-edit-rmbg"}
    if not compatible_edit_aliases.intersection(model_config.aliases):
        parser.error(
            "mflux-generate-fibo-edit requires a FIBO Edit model; metadata or --model resolved to an incompatible model."
        )

    return model_config


def _is_rmbg(model_config: ModelConfig) -> bool:
    return "fibo-edit-rmbg" in model_config.aliases


def _validate_matte_output(parser: CommandLineParser, args, model_config: ModelConfig) -> None:
    if args.matte_output is not None and not _is_rmbg(model_config):
        parser.error("--matte-output is only supported with --model fibo-edit-rmbg.")


def _apply_default_guidance(args, model_config: ModelConfig) -> None:
    if args.guidance is None:
        args.guidance = FIBO_EDIT_RMBG_GUIDANCE_DEFAULT if _is_rmbg(model_config) else ui_defaults.GUIDANCE_SCALE


def _json_prompt_for_edit(args, model_config: ModelConfig) -> str:
    default_if_missing = FIBO_EDIT_RMBG_DEFAULT_JSON_PROMPT if _is_rmbg(model_config) else None
    return FiboEditUtil.get_json_prompt_for_edit(
        args,
        quantize=args.quantize,
        default_json_prompt_if_missing=default_if_missing,
    )


def _save_edit_result(
    image: GeneratedImage,
    args,
    model_config: ModelConfig,
    seed: int,
) -> None:
    out_path = args.output.format(seed=seed)
    if _is_rmbg(model_config):
        rgba_pil = FiboEditUtil.build_rgba_composite_image(args.image_path, image.image)
        image.save_with_image(
            path=out_path,
            pixel_image=rgba_pil,
            export_json_metadata=args.metadata,
        )
        if args.matte_output is not None:
            image.save(
                path=args.matte_output.format(seed=seed),
                export_json_metadata=False,
            )
    else:
        image.save(path=out_path, export_json_metadata=args.metadata)


def main():
    parser = CommandLineParser(description="Generate an edited image using Bria FIBO Edit.")
    parser.add_general_arguments()
    parser.add_model_arguments(require_model_arg=False)
    parser.set_defaults(model="fibo-edit")
    parser.add_lora_arguments()
    parser.add_image_generator_arguments(
        supports_metadata_config=True,
        require_prompt=False,
        supports_dimension_scale_factor=True,
    )
    parser.add_argument("--image-path", type=Path, required=False, help="Local path to source image for editing.")
    parser.add_argument("--mask-path", type=Path, default=None, help="Optional mask image path for localized edits.")
    parser.add_argument("--matte-output", type=str, default=None, help="fibo-edit-rmbg only: also save the raw grayscale matte. Supports {seed} like --output.")  # fmt: skip
    parser.add_output_arguments()
    args = parser.parse_args()

    if args.image_path is None:
        parser.error("--image-path is required, or 'image_path' must be specified in the config file.")

    model_config = _resolve_fibo_edit_model_config(parser, args)
    _validate_matte_output(parser, args, model_config)
    _apply_default_guidance(args, model_config)
    json_prompt = _json_prompt_for_edit(args, model_config)

    fibo_edit = FIBOEdit(
        quantize=args.quantize,
        model_path=args.model_path,
        lora_paths=args.lora_paths,
        lora_scales=args.lora_scales,
        model_config=model_config,
    )

    memory_saver = CallbackManager.register_callbacks(
        args=args,
        model=fibo_edit,
        latent_creator=FiboLatentCreator,
    )

    try:
        width, height = DimensionResolver.resolve(
            width=args.width,
            height=args.height,
            reference_image_path=args.image_path,
        )
        for seed in args.seed:
            image = fibo_edit.generate_image(
                seed=seed,
                prompt=json_prompt,
                image_path=args.image_path,
                mask_path=args.mask_path,
                width=width,
                height=height,
                guidance=args.guidance,
                num_inference_steps=args.steps,
                scheduler="flow_match_euler_discrete",
                negative_prompt=PromptUtil.read_negative_prompt(args),
            )
            _save_edit_result(image, args, model_config, seed)
    except (StopImageGenerationException, PromptFileReadError, ValueError) as exc:
        print(exc)
    finally:
        if memory_saver:
            print(memory_saver.memory_stats())


if __name__ == "__main__":
    main()



import math

import mlx.core as mx
from mlx import nn


class Flux2TimestepGuidanceEmbeddings(nn.Module):
    def __init__(self, in_channels: int = 256, embedding_dim: int = 6144, guidance_embeds: bool = True):
        super().__init__()
        self.in_channels = in_channels
        self.embedding_dim = embedding_dim
        self.guidance_embeds = guidance_embeds
        self.linear_1 = nn.Linear(in_channels, embedding_dim, bias=False)
        self.linear_2 = nn.Linear(embedding_dim, embedding_dim, bias=False)
        self.guidance_linear_1 = nn.Linear(in_channels, embedding_dim, bias=False) if guidance_embeds else None
        self.guidance_linear_2 = nn.Linear(embedding_dim, embedding_dim, bias=False) if guidance_embeds else None

    def __call__(self, timestep: mx.array, guidance: mx.array | None) -> mx.array:
        timestep = Flux2TimestepGuidanceEmbeddings._timestep_embedding(timestep.astype(mx.float32), self.in_channels)
        timesteps_emb = self.linear_2(nn.silu(self.linear_1(timestep)))
        if guidance is not None and self.guidance_linear_1 is not None and self.guidance_linear_2 is not None:
            guidance = Flux2TimestepGuidanceEmbeddings._timestep_embedding(
                guidance.astype(mx.float32), self.in_channels
            )
            guidance_emb = self.guidance_linear_2(nn.silu(self.guidance_linear_1(guidance)))
            return timesteps_emb + guidance_emb
        return timesteps_emb

    @staticmethod
    def _timestep_embedding(timesteps: mx.array, dim: int, flip_sin_to_cos: bool = True) -> mx.array:
        half = dim // 2
        freqs = mx.exp(-math.log(10000.0) * mx.arange(0, half, dtype=mx.float32) / half)
        args = timesteps[:, None] * freqs[None, :]
        emb = mx.concatenate([mx.sin(args), mx.cos(args)], axis=-1)
        if flip_sin_to_cos:
            emb = mx.concatenate([emb[:, half:], emb[:, :half]], axis=-1)
        if dim % 2 == 1:
            emb = mx.concatenate([emb, mx.zeros((emb.shape[0], 1), dtype=emb.dtype)], axis=-1)
        return emb



from mflux.models.common.training.lora.path_util import (
    expand_module_paths,
    get_at_path,
    set_at_path,
)
from mflux.models.common.training.lora.target_injector import inject_lora_targets

__all__ = ["get_at_path", "set_at_path", "expand_module_paths", "inject_lora_targets"]



import mlx.core as mx
from mlx import nn


class Flux2Upsample2D(nn.Module):
    def __init__(self, channels: int, out_channels: int | None = None):
        super().__init__()
        out_channels = out_channels or channels
        self.conv = nn.Conv2d(channels, out_channels, kernel_size=3, stride=1, padding=1)

    def __call__(self, hidden_states: mx.array) -> mx.array:
        hidden_states = mx.repeat(hidden_states, 2, axis=2)
        hidden_states = mx.repeat(hidden_states, 2, axis=3)
        hidden_states = mx.transpose(hidden_states, (0, 2, 3, 1))
        hidden_states = self.conv(hidden_states)
        return mx.transpose(hidden_states, (0, 3, 1, 2))



from __future__ import annotations

import datetime
import json
import random
from pathlib import Path
from typing import Any

from mflux.models.common.training.dataset.batch import Batch
from mflux.models.common.training.dataset.dataset import Dataset
from mflux.models.common.training.state.training_spec import TrainingSpec
from mflux.models.common.training.state.zip_util import ZipUtil


class Iterator:
    def __init__(
        self,
        dataset: Dataset,
        batch_size: int,
        num_epochs: int | None = None,
        seed: int | None = None,
        position: int = 0,
        epoch: int = 0,
        num_iterations: int = 0,
        current_permutation: list[int] | None = None,
        rng_state: tuple | None = None,
        start_date_time: datetime.datetime | None = None,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.total_items = dataset.size()
        self.num_epochs = num_epochs
        self.seed = seed
        self.rng = random.Random(seed)

        self._position = position
        self._epoch = epoch
        self.num_iterations = num_iterations
        self.start_date_time = start_date_time or datetime.datetime.now()

        if rng_state is not None:
            self.rng.setstate(rng_state)

        if current_permutation is not None:
            self._current_permutation = current_permutation
        else:
            self._initialize_permutation()

    @staticmethod
    def from_spec(training_spec: TrainingSpec, dataset: Dataset) -> "Iterator":
        if training_spec.training_loop.iterator_state_path is not None:
            return Iterator.from_json(training_spec=training_spec, iterator_path=training_spec.training_loop.iterator_state_path, dataset=dataset)  # fmt: off

        return Iterator(
            seed=training_spec.seed,
            dataset=dataset,
            batch_size=training_spec.training_loop.batch_size,
            num_epochs=training_spec.training_loop.num_epochs,
        )

    def _initialize_permutation(self):
        self._current_permutation = list(range(self.total_items))
        self.rng.shuffle(self._current_permutation)

    def __iter__(self):
        return self

    def __next__(self):
        if self.num_epochs is not None and self._epoch >= self.num_epochs:
            raise StopIteration()

        if self._position >= self.total_items:
            self._position = 0
            self._epoch += 1
            if self.num_epochs is not None and self._epoch >= self.num_epochs:
                raise StopIteration()
            self._initialize_permutation()

        remaining = self.total_items - self._position
        current_batch_size = min(self.batch_size, remaining)
        batch_indices = self._current_permutation[self._position : self._position + current_batch_size]
        self._position += current_batch_size
        self.num_iterations += 1

        data_items = [self.dataset.data[i] for i in batch_indices]
        return Batch(data=data_items, rng=self.rng)

    def to_dict(self) -> dict[str, Any]:
        return {
            "position": self._position,
            "epoch": self._epoch,
            "num_iterations": self.num_iterations,
            "current_permutation": self._current_permutation,
            "rng_state": list(self.rng.getstate()),
            "batch_size": self.batch_size,
            "num_epochs": self.num_epochs,
            "seed": self.seed,
            "start_date_time": self.start_date_time.strftime("%Y-%m-%d %H:%M:%S"),
        }

    @classmethod
    def from_dict(cls, state_dict: dict[str, Any], dataset: Dataset) -> "Iterator":
        rng_state = state_dict["rng_state"]
        rng_state = (rng_state[0], tuple(rng_state[1]), rng_state[2])
        return cls(
            dataset=dataset,
            batch_size=state_dict["batch_size"],
            num_epochs=state_dict["num_epochs"],
            seed=state_dict.get("seed"),
            position=state_dict["position"],
            epoch=state_dict["epoch"],
            num_iterations=state_dict["num_iterations"],
            current_permutation=state_dict["current_permutation"],
            rng_state=rng_state,
            start_date_time=datetime.datetime.strptime(state_dict["start_date_time"], "%Y-%m-%d %H:%M:%S"),
        )

    @classmethod
    def from_json(cls, training_spec: TrainingSpec, iterator_path: str, dataset: Dataset) -> "Iterator":
        data = ZipUtil.unzip(
            zip_path=training_spec.checkpoint_path,
            filename=iterator_path,
            loader=lambda x: json.load(open(x, "r")),
        )
        if "seed" not in data:
            data["seed"] = training_spec.seed
        return cls.from_dict(data, dataset)

    def save(self, path: Path) -> None:
        with open(path, "w") as f:
            json.dump(self.to_dict(), f, indent=4)

    def get_validation_batch(self) -> Batch:
        data_items = self.dataset.data
        total = len(data_items)
        if total > 10:
            # Deterministic sample without affecting training RNG.
            base_seed = self.seed or 0
            validation_rng = random.Random(base_seed + 1337 + self.num_iterations)
            indices = validation_rng.sample(range(total), k=10)
            data_items = [data_items[i] for i in indices]
        else:
            validation_rng = random.Random(self.seed or 0)
        return Batch(data=data_items, rng=validation_rng)

    def total_number_of_steps(self) -> int:
        total_epochs = self.num_epochs  # type: ignore[assignment]
        if total_epochs is None:
            return 0
        batches_per_epoch = (self.total_items + self.batch_size - 1) // self.batch_size
        return int(batches_per_epoch * total_epochs)



from mflux.models.common.training.dataset.batch import Batch, DataItem
from mflux.models.common.training.dataset.dataset import Dataset
from mflux.models.common.training.dataset.iterator import Iterator

__all__ = ["Batch", "DataItem", "Dataset", "Iterator"]



import math
from functools import partial
from typing import TYPE_CHECKING

import mlx.core as mx

if TYPE_CHECKING:
    from mflux.models.common.config.config import Config

from mflux.models.common.schedulers.base_scheduler import BaseScheduler


@partial(mx.compile, shapeless=True)
def _step(noise, latents, s1, s2):
    dt = (s1 - s2).astype(latents.dtype)
    noise = noise.astype(latents.dtype)
    return latents + dt * noise


class FlowMatchEulerDiscreteScheduler(BaseScheduler):
    def __init__(self, config: "Config"):
        self.config = config
        self.model_config = config.model_config
        self.num_train_timesteps = 1000
        self.shift_terminal = 0.02
        self._sigmas, self._timesteps = self._compute_timesteps_and_sigmas()

    @property
    def sigmas(self) -> mx.array:
        return self._sigmas

    @property
    def timesteps(self) -> mx.array:
        return self._timesteps

    def set_image_seq_len(self, image_seq_len: int) -> None:
        self._timesteps, self._sigmas = FlowMatchEulerDiscreteScheduler.get_timesteps_and_sigmas(
            image_seq_len=image_seq_len,
            num_inference_steps=self.config.num_inference_steps,
        )

    def set_mu(self, mu: float) -> None:
        num_steps = self.config.num_inference_steps
        sigmas = mx.linspace(1.0, 1.0 / num_steps, num_steps, dtype=mx.float32)
        sigmas = FlowMatchEulerDiscreteScheduler._time_shift_exponential_array(mu, 1.0, sigmas)
        self._timesteps = sigmas * self.num_train_timesteps
        sigmas = mx.concatenate([sigmas, mx.zeros((1,), dtype=sigmas.dtype)], axis=0)
        self._sigmas = sigmas

    @staticmethod
    def _compute_linear_mu(
        image_seq_len: int,
        base_seq_len: int = 256,
        max_seq_len: int = 4096,
        base_shift: float = 0.5,
        max_shift: float = 1.15,
    ) -> float:
        m = (max_shift - base_shift) / (max_seq_len - base_seq_len)
        b = base_shift - m * base_seq_len
        return float(m * image_seq_len + b)

    @staticmethod
    def get_timesteps_and_sigmas(
        image_seq_len: int,
        num_inference_steps: int,
        num_train_timesteps: int = 1000,
    ) -> tuple[mx.array, mx.array]:
        sigmas = mx.linspace(1.0, 1.0 / num_inference_steps, num_inference_steps, dtype=mx.float32)
        mu = FlowMatchEulerDiscreteScheduler._compute_empirical_mu(
            image_seq_len=image_seq_len,
            num_steps=num_inference_steps,
        )
        sigmas = FlowMatchEulerDiscreteScheduler._time_shift_exponential_array(mu, 1.0, sigmas)
        timesteps = sigmas * num_train_timesteps
        sigmas = mx.concatenate([sigmas, mx.zeros((1,), dtype=sigmas.dtype)], axis=0)
        return timesteps, sigmas

    @staticmethod
    def _compute_empirical_mu(image_seq_len: int, num_steps: int) -> float:
        a1, b1 = 8.73809524e-05, 1.89833333
        a2, b2 = 0.00016927, 0.45666666
        if image_seq_len > 4300:
            return float(a2 * image_seq_len + b2)
        m_200 = a2 * image_seq_len + b2
        m_10 = a1 * image_seq_len + b1
        a = (m_200 - m_10) / 190.0
        b = m_200 - 200.0 * a
        return float(a * num_steps + b)

    @staticmethod
    def _time_shift_exponential(mu: float, sigma_power: float, t: float) -> float:
        return math.exp(mu) / (math.exp(mu) + ((1.0 / t - 1.0) ** sigma_power))

    @staticmethod
    def _time_shift_exponential_array(mu: float, sigma_power: float, t: mx.array) -> mx.array:
        return mx.exp(mu) / (mx.exp(mu) + ((1.0 / t - 1.0) ** sigma_power))

    def _stretch_to_terminal(self, sigmas: list[float]) -> list[float]:
        one_minus_sigmas = [1.0 - s for s in sigmas]
        scale_factor = one_minus_sigmas[-1] / (1.0 - self.shift_terminal)
        stretched = [1.0 - (oms / scale_factor) for oms in one_minus_sigmas]
        return stretched

    def _compute_timesteps_and_sigmas(self) -> tuple[mx.array, mx.array]:
        num_steps = self.config.num_inference_steps
        sigma_min = 1.0 / self.num_train_timesteps
        sigma_max = 1.0
        timesteps_linear = [
            sigma_max * self.num_train_timesteps
            - i * (sigma_max - sigma_min) * self.num_train_timesteps / (num_steps - 1)
            for i in range(num_steps)
        ]
        sigmas_linear = [t / self.num_train_timesteps for t in timesteps_linear]
        sigmas_shifted = [FlowMatchEulerDiscreteScheduler._time_shift_exponential(1.0, 1.0, s) for s in sigmas_linear]
        sigmas_final = self._stretch_to_terminal(sigmas_shifted)
        timesteps = [s * self.num_train_timesteps for s in sigmas_final]
        sigmas_with_zero = sigmas_final + [0.0]
        sigmas_arr = mx.array(sigmas_with_zero, dtype=mx.float32)
        timesteps_arr = mx.array(timesteps, dtype=mx.float32)
        return sigmas_arr, timesteps_arr

    def step(self, noise: mx.array, timestep: int, latents: mx.array, **kwargs) -> mx.array:
        sigmas = kwargs.get("sigmas", self._sigmas)
        return _step(noise, latents, sigmas[timestep + 1], sigmas[timestep])

    def scale_model_input(self, latents: mx.array, t: int) -> mx.array:
        return latents



from pathlib import Path
from typing import TYPE_CHECKING

import transformers
from huggingface_hub import snapshot_download
from huggingface_hub.utils import LocalEntryNotFoundError

from mflux.models.common.tokenizer.tokenizer import (
    BaseTokenizer,
    LanguageTokenizer,
    VisionLanguageTokenizer,
)

if TYPE_CHECKING:
    from mflux.models.common.weights.loading.weight_definition import TokenizerDefinition


class TokenizerLoader:
    @staticmethod
    def load(
        definition: "TokenizerDefinition",
        model_path: str,
    ) -> BaseTokenizer:
        tokenizer_path = TokenizerLoader._resolve_path(
            model_path=model_path,
            hf_subdir=definition.hf_subdir,
            fallback_subdirs=definition.fallback_subdirs,
            download_patterns=definition.download_patterns,
            tokenizer_class=definition.tokenizer_class,
        )

        raw_tokenizer = TokenizerLoader._load_raw_tokenizer(
            tokenizer_path=tokenizer_path,
            tokenizer_class=definition.tokenizer_class,
            chat_template=definition.chat_template,
        )

        return TokenizerLoader._create_tokenizer(
            raw_tokenizer=raw_tokenizer,
            definition=definition,
        )

    @staticmethod
    def load_all(
        definitions: list["TokenizerDefinition"],
        model_path: str,
        max_length_overrides: dict[str, int] | None = None,
    ) -> dict[str, BaseTokenizer]:
        max_length_overrides = max_length_overrides or {}
        result = {}
        for d in definitions:
            tokenizer = TokenizerLoader.load(d, model_path)
            if d.name in max_length_overrides:
                tokenizer.max_length = max_length_overrides[d.name]
            result[d.name] = tokenizer
        return result

    @staticmethod
    def _resolve_path(
        model_path: str,
        hf_subdir: str,
        fallback_subdirs: list[str] | None,
        download_patterns: list[str] | None,
        tokenizer_class: str,
    ) -> Path:
        # Resolution algorithm:
        # 1. Resolve a local path or an initial Hugging Face snapshot root.
        # 2. Prefer the primary tokenizer path and preserve real load failures.
        # 3. For cached HF snapshots, keep any loadable fallback layout before refreshing or failing.
        # 4. If a cached HF snapshot is incomplete, refresh it once over the network.
        # 5. Only then raise a targeted missing-tokenizer or unloadable-tokenizer error.
        patterns = download_patterns or [f"{hf_subdir}/**"]
        root_path, is_hf_repo, saw_cached_snapshot, download_error = TokenizerLoader._resolve_root_path(
            model_path=model_path,
            patterns=patterns,
        )

        if root_path is None:
            TokenizerLoader._raise_missing_tokenizer_error(
                model_path=model_path,
                hf_subdir=hf_subdir,
                fallback_subdirs=fallback_subdirs,
                is_hf_repo=is_hf_repo,
                saw_cached_snapshot=saw_cached_snapshot,
                download_error=download_error,
            )
        assert root_path is not None

        tokenizer_path, primary_load_error, primary_has_tokenizer_artifacts = TokenizerLoader._inspect_primary_path(
            root_path=root_path,
            hf_subdir=hf_subdir,
            tokenizer_class=tokenizer_class,
        )
        if tokenizer_path is not None:
            return tokenizer_path

        allow_cached_fallback = is_hf_repo and saw_cached_snapshot

        # If an alternate cached layout already works, keep the offline-first behavior.
        if allow_cached_fallback or not primary_has_tokenizer_artifacts:
            fallback_path = TokenizerLoader._resolve_fallback_path(root_path, fallback_subdirs, tokenizer_class)
            if fallback_path is not None:
                return fallback_path

        if is_hf_repo and saw_cached_snapshot:
            refreshed_root_path, download_error = TokenizerLoader._refresh_hf_root_path(
                model_path=model_path,
                patterns=patterns,
            )
            if refreshed_root_path is not None:
                root_path = refreshed_root_path
                tokenizer_path, primary_load_error, primary_has_tokenizer_artifacts = (
                    TokenizerLoader._inspect_primary_path(
                        root_path=root_path,
                        hf_subdir=hf_subdir,
                        tokenizer_class=tokenizer_class,
                    )
                )
                if tokenizer_path is not None:
                    return tokenizer_path

                if allow_cached_fallback or not primary_has_tokenizer_artifacts:
                    fallback_path = TokenizerLoader._resolve_fallback_path(root_path, fallback_subdirs, tokenizer_class)
                    if fallback_path is not None:
                        return fallback_path

        if primary_load_error is not None and primary_has_tokenizer_artifacts and download_error is None:
            TokenizerLoader._raise_unloadable_tokenizer_error(
                model_path=model_path,
                tokenizer_path=root_path / hf_subdir,
                is_hf_repo=is_hf_repo,
                load_error=primary_load_error,
            )

        TokenizerLoader._raise_missing_tokenizer_error(
            model_path=model_path,
            hf_subdir=hf_subdir,
            fallback_subdirs=fallback_subdirs,
            is_hf_repo=is_hf_repo,
            saw_cached_snapshot=saw_cached_snapshot,
            download_error=download_error,
        )

    @staticmethod
    def _resolve_root_path(
        model_path: str,
        patterns: list[str],
    ) -> tuple[Path | None, bool, bool, Exception | None]:
        expanded = Path(model_path).expanduser()
        if expanded.exists():
            return expanded, False, False, None
        if TokenizerLoader._is_huggingface_repo_id(model_path):
            root_path, saw_cached_snapshot, download_error = TokenizerLoader._resolve_hf_root_path(
                model_path=model_path,
                patterns=patterns,
            )
            return root_path, True, saw_cached_snapshot, download_error
        raise FileNotFoundError(
            f"Model not found: '{model_path}'. "
            f"If local path, make sure it exists. "
            f"If HuggingFace repo, use 'org/model' format."
        )

    @staticmethod
    def _is_huggingface_repo_id(model_path: str) -> bool:
        return "/" in model_path and model_path.count("/") == 1 and not model_path.startswith(("./", "../", "~/"))

    @staticmethod
    def _resolve_hf_root_path(
        model_path: str,
        patterns: list[str],
    ) -> tuple[Path | None, bool, Exception | None]:
        try:
            root_path = Path(
                snapshot_download(
                    repo_id=model_path,
                    allow_patterns=patterns,
                    local_files_only=True,
                )
            )
        except LocalEntryNotFoundError:
            root_path, download_error = TokenizerLoader._download_hf_root_path(
                model_path=model_path,
                patterns=patterns,
            )
            return root_path, False, download_error

        return root_path, True, None

    @staticmethod
    def _download_hf_root_path(
        model_path: str,
        patterns: list[str],
    ) -> tuple[Path | None, Exception | None]:
        try:
            root_path = Path(
                snapshot_download(
                    repo_id=model_path,
                    allow_patterns=patterns,
                )
            )
        except Exception as exc:  # noqa: BLE001
            return None, exc
        return root_path, None

    @staticmethod
    def _refresh_hf_root_path(
        model_path: str,
        patterns: list[str],
    ) -> tuple[Path | None, Exception | None]:
        return TokenizerLoader._download_hf_root_path(
            model_path=model_path,
            patterns=patterns,
        )

    @staticmethod
    def _inspect_primary_path(
        root_path: Path,
        hf_subdir: str,
        tokenizer_class: str,
    ) -> tuple[Path | None, Exception | None, bool]:
        tokenizer_path = root_path / hf_subdir
        is_primary_loadable, primary_load_error = TokenizerLoader._probe_tokenizer_path(
            tokenizer_path,
            tokenizer_class,
        )
        if is_primary_loadable:
            return tokenizer_path, None, True

        primary_has_tokenizer_artifacts = TokenizerLoader._has_tokenizer_artifacts(tokenizer_path)
        return None, primary_load_error, primary_has_tokenizer_artifacts

    @staticmethod
    def _is_tokenizer_loadable(path: Path, tokenizer_class: str) -> bool:
        is_loadable, _ = TokenizerLoader._probe_tokenizer_path(path, tokenizer_class)
        return is_loadable

    @staticmethod
    def _probe_tokenizer_path(path: Path, tokenizer_class: str) -> tuple[bool, Exception | None]:
        if not path.exists():
            return False, None

        TokenizerLoader._get_tokenizer_class(tokenizer_class)

        try:
            TokenizerLoader._load_raw_tokenizer(
                tokenizer_path=path,
                tokenizer_class=tokenizer_class,
            )
        except Exception as exc:  # noqa: BLE001
            return False, exc
        return True, None

    @staticmethod
    def _has_tokenizer_artifacts(path: Path) -> bool:
        if not path.exists():
            return False

        # Broad markers used only to distinguish an alternate layout from a broken tokenizer.
        tokenizer_artifacts = [
            "added_tokens.json",
            "merges.txt",
            "sentencepiece.bpe.model",
            "special_tokens_map.json",
            "spiece.model",
            "tokenizer.json",
            "tokenizer.model",
            "tokenizer_config.json",
            "vocab.json",
            "vocab.txt",
        ]
        return any((path / filename).exists() for filename in tokenizer_artifacts)

    @staticmethod
    def _resolve_fallback_path(
        root_path: Path,
        fallback_subdirs: list[str] | None,
        tokenizer_class: str,
    ) -> Path | None:
        if not fallback_subdirs:
            return None

        for subdir in fallback_subdirs:
            fallback_path = root_path if subdir == "." else root_path / subdir
            if TokenizerLoader._is_tokenizer_loadable(fallback_path, tokenizer_class):
                return fallback_path

        return None

    @staticmethod
    def _raise_missing_tokenizer_error(
        model_path: str,
        hf_subdir: str,
        fallback_subdirs: list[str] | None,
        is_hf_repo: bool,
        saw_cached_snapshot: bool,
        download_error: Exception | None = None,
    ) -> None:
        primary_location = "repository root" if hf_subdir in ("", ".") else hf_subdir
        checked_locations = [repr(primary_location)]
        if fallback_subdirs:
            checked_locations.extend(
                repr("repository root" if subdir in ("", ".") else subdir) for subdir in fallback_subdirs
            )
        checked_summary = ", ".join(checked_locations)

        if is_hf_repo:
            if saw_cached_snapshot and download_error is not None:
                message = (
                    f"Incomplete Hugging Face tokenizer cache for '{model_path}'. "
                    f"No usable tokenizer files were found in {checked_summary}. "
                    f"Re-run with network access or clear/redownload the cache."
                )
            elif download_error is not None:
                message = (
                    f"Failed to download tokenizer files for Hugging Face repo '{model_path}'. "
                    f"Checked {checked_summary}."
                )
            else:
                message = (
                    f"No usable tokenizer files were found for Hugging Face repo '{model_path}'. "
                    f"Checked {checked_summary}."
                )
        else:
            message = (
                f"No usable tokenizer files were found for local model path '{model_path}'. Checked {checked_summary}."
            )

        if download_error is not None:
            raise FileNotFoundError(message) from download_error
        raise FileNotFoundError(message)

    @staticmethod
    def _raise_unloadable_tokenizer_error(
        model_path: str,
        tokenizer_path: Path,
        is_hf_repo: bool,
        load_error: Exception,
    ) -> None:
        location_type = "Hugging Face repo" if is_hf_repo else "local model path"
        raise RuntimeError(
            f"Found tokenizer files for {location_type} '{model_path}' at '{tokenizer_path}', but failed to load them."
        ) from load_error

    @staticmethod
    def _load_raw_tokenizer(
        tokenizer_path: Path,
        tokenizer_class: str,
        chat_template: str | None = None,
    ):
        cls = TokenizerLoader._get_tokenizer_class(tokenizer_class)

        # Workaround for Qwen2Tokenizer bug in transformers 5.0.0rc0
        # The tokenizer doesn't properly load vocab/merges from files, need to pass them directly
        if tokenizer_class == "Qwen2Tokenizer":
            tokenizer = TokenizerLoader._load_qwen2_tokenizer_workaround(tokenizer_path, cls)
        else:
            tokenizer = cls.from_pretrained(
                pretrained_model_name_or_path=str(tokenizer_path),
                local_files_only=True,
            )

        # Apply chat_template from definition if provided and not already set
        if chat_template and not getattr(tokenizer, "chat_template", None):
            tokenizer.chat_template = chat_template

        return tokenizer

    @staticmethod
    def _get_tokenizer_class(tokenizer_class: str):
        if hasattr(transformers, tokenizer_class):
            return getattr(transformers, tokenizer_class)
        raise ValueError(f"Unknown tokenizer class: {tokenizer_class}")

    @staticmethod
    def _load_qwen2_tokenizer_workaround(tokenizer_path: Path, cls):
        import json

        from tokenizers import AddedToken

        vocab_file = tokenizer_path / "vocab.json"
        merges_file = tokenizer_path / "merges.txt"
        config_file = tokenizer_path / "tokenizer_config.json"

        if not vocab_file.exists() or not merges_file.exists():
            # Fall back to standard loading if files don't exist
            return cls.from_pretrained(
                pretrained_model_name_or_path=str(tokenizer_path),
                local_files_only=True,
            )

        # Load vocab as dict[str, int]
        with open(vocab_file, encoding="utf-8") as f:
            vocab = json.load(f)

        # Load merges as list of tuples (pair of strings)
        # The merges.txt file has format: "token1 token2" per line (skip header if present)
        merges = []
        with open(merges_file, encoding="utf-8") as f:
            for line in f:
                line = line.rstrip()
                # Skip empty lines and the optional version header (e.g., "#version: 0.2")
                if not line or line.startswith("#"):
                    continue
                parts = line.split()
                if len(parts) == 2:
                    merges.append((parts[0], parts[1]))

        # Load tokenizer config for special tokens and chat template
        config_kwargs = {}
        chat_template = None
        if config_file.exists():
            with open(config_file, encoding="utf-8") as f:
                config = json.load(f)

            # Extract added_tokens_decoder to add special tokens
            added_tokens_decoder = config.get("added_tokens_decoder", {})
            if added_tokens_decoder:
                # Convert to the format expected by the tokenizer
                config_kwargs["added_tokens_decoder"] = {
                    int(k): AddedToken(
                        content=v["content"],
                        lstrip=v.get("lstrip", False),
                        rstrip=v.get("rstrip", False),
                        single_word=v.get("single_word", False),
                        normalized=v.get("normalized", False),
                        special=v.get("special", True),
                    )
                    for k, v in added_tokens_decoder.items()
                }

            # Extract chat_template if present
            chat_template = config.get("chat_template")

        tokenizer = cls(vocab=vocab, merges=merges, **config_kwargs)

        # Set chat_template after initialization (not a constructor param)
        if chat_template:
            tokenizer.chat_template = chat_template

        return tokenizer

    @staticmethod
    def _create_tokenizer(
        raw_tokenizer,
        definition: "TokenizerDefinition",
    ) -> BaseTokenizer:
        encoder_class = definition.encoder_class

        if encoder_class is VisionLanguageTokenizer:
            if definition.processor_class is None:
                raise ValueError("VisionLanguageTokenizer requires processor_class in definition")
            processor = definition.processor_class(tokenizer=raw_tokenizer)
            return VisionLanguageTokenizer(
                tokenizer=raw_tokenizer,
                processor=processor,
                max_length=definition.max_length,
                template=definition.template,
                image_token=definition.image_token,
            )
        else:
            # Default to LanguageTokenizer for all text-only cases
            return LanguageTokenizer(
                tokenizer=raw_tokenizer,
                max_length=definition.max_length,
                padding=definition.padding,
                template=definition.template,
                use_chat_template=definition.use_chat_template,
                chat_template_kwargs=definition.chat_template_kwargs or {},
                add_special_tokens=definition.add_special_tokens,
            )
